//**********************************************************************
//  aipGood.cxx   -   Member functions for aipGood.h
//
//  aipGood.h implements goodness, importance and likelihood.
//
//  Copyright Brian Marshall 1999
//
//  See the license at end of this file.
//
//  99/03/07  Brian Marshall  development begun
//  99/05/01  Brian Marshall  naming convention changed
//
//----------------------------------------------------------------------

#include "aipGood.h"

//======================================================================
//  aipGBase.
//
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
//  Constructors and destructor.

aipGBase::aipGBase  (void)               { m_val = 0;       }

aipGBase::aipGBase  (const long x)       { m_val = x;       }

aipGBase::aipGBase  (const aipGBase& x)  { m_val = x.val(); }

aipGBase::~aipGBase (void) {}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
//  Assignment.

aipGBase& aipGBase::operator = (const aipGBase& x) { 
  m_val = x.val(); 
  return *this;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
//  Comparison.

bool aipGBase::operator <  (const aipGBase& x) const { 
  return m_val <  x.val(); 
}

bool aipGBase::operator >  (const aipGBase& x) const { 
  return m_val >  x.val(); 
}

bool aipGBase::operator == (const aipGBase& x) const { 
  return m_val == x.val(); 
}

bool aipGBase::operator != (const aipGBase& x) const { 
  return m_val != x.val(); 
}

bool aipGBase::operator <= (const aipGBase& x) const { 
  return m_val <= x.val(); 
}

bool aipGBase::operator >= (const aipGBase& x) const { 
  return m_val >= x.val(); 
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
//  Internal representation retrieval and updating.

long aipGBase::val (void) const { 
  return m_val; 
}

void aipGBase::set_val (const long x) { 
  m_val = x; 
}


//======================================================================
//  aipGoodness   -   How good or bad something is.
//
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
//  Combine two aipGoodness objects.

aipGoodness aipGoodness::operator +  (const aipGoodness& x) const { 
  return aipGoodness (m_val + x.val()); 
}

aipGoodness& aipGoodness::operator += (const aipGoodness& x) { 
  m_val += x.val();
  return *this;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
//  Remove the effect of a aipGoodness from a combination.

aipGoodness aipGoodness::operator -  (const aipGoodness& x) const { 
  return aipGoodness (m_val - x.val()); 
}

aipGoodness& aipGoodness::operator -= (const aipGoodness& x) { 
  m_val -= x.val(); 
  return *this;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
//  Return a description of this aipGoodness.

const char * aipGoodness::description (void) const {

  if (m_val <= aipBad.val()) {

    if (m_val <= aipForbidden.val()) {
      return "Forbidden";
    } else if (m_val <= aipVeryVeryBad.val()) {
      return "Very Very Bad";
    } else if (m_val <= aipVeryBad.val()) {
      return "Very Bad";
    } else if (m_val <= aipQuiteBad.val()) {
      return "Quite Bad";
    } else {
      return "Bad";
    }

  } else if (m_val <= aipVerySlightlyBad.val()) {

    if (m_val <= aipFairlyBad.val()) {
      return "Fairly Bad";
    } else if (m_val <= aipSomewhatBad.val()) {
      return "Somewhat Bad";
    } else if (m_val <= aipSlightlyBad.val()) {
      return "Slightly Bad";
    } else {
      return "Very Slightly Bad";
    }
    
  } else if (m_val >= aipGood.val()) {
      
    if (m_val >= aipManditory.val()) {
      return "Manditory";
    } else if (m_val >= aipVeryVeryGood.val()) {
      return "Very Very Good";
    } else if (m_val >= aipVeryGood.val()) {
      return "Very Good";
    } else if (m_val >= aipQuiteGood.val()) {
      return "Quite Good";
    } else {
      return "Good";
    }

  } else if (m_val >= aipVerySlightlyGood.val()) {

    if (m_val >= aipFairlyGood.val()) {
      return "Fairly Good";
    } else if (m_val >= aipSomewhatGood.val()) {
      return "Somewhat Good";
    } else if (m_val >= aipSlightlyGood.val()) {
      return "Slightly Good";
    } else {
      return "Very Slightly Good";
    }
    
  }

  return "Neutral";

}

//======================================================================
//  Static functions to apply and unapply a qualifier.

const long aipDenominator = 128;

static long apply (const long a, const long b) {
  return (a * b) / aipDenominator;
}

static long unapply (const long a, const long b) {
  if (b == 0) return a;
  return (a * aipDenominator) / b; 
}


//======================================================================
//  aipImportance   -   How important something is.
//
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
//  Return a description of this aipImportance.

const char * aipImportance::description (void) const {

  if (m_val >= aipVeryVeryImportant.val()) {
    return "Very Very Important";
  } else if (m_val >= aipVeryImportant.val()) {
    return "Very Important";
  } else if (m_val >= aipQuiteImportant.val()) {
    return "Quite Important";
  } else if (m_val >= aipImportant.val()){
    return "Important";
  } else if (m_val >= aipFairlyImportant.val()) {
    return "Fairly Important";
  } else if (m_val >= aipSomewhatImportant.val()) {
    return "Somewhat Important";
  } else if (m_val >= aipSlightlyImportant.val()) {
    return "Slightly Important";
  } else if (m_val >= aipVerySlightlyImportant.val()) {
    return "Very Slightly Important";
  }
  
  return "No Importance";

}

//======================================================================
//  aipLikelihood   -   How likely something is to happen.
//
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
//  Return a description of this aipLikelihood.

const char * aipLikelihood::description (void) const {

  if (m_val >= aipVeryVeryLikely.val()) {
    return "Very Very Likely";
  } else if (m_val >= aipVeryLikely.val()) {
    return "Very Likely";
  } else if (m_val >= aipQuiteLikely.val()) {
    return "Quite Likely";
  } else if (m_val >= aipLikely.val()){
    return "Likely";
  } else if (m_val >= aipFairlyLikely.val()) {
    return "Fairly Likely";
  } else if (m_val >= aipSomewhatLikely.val()) {
    return "Somewhat Likely";
  } else if (m_val >= aipSlightlyLikely.val()) {
    return "Slightly Likely";
  } else if (m_val >= aipVerySlightlyLikely.val()) {
    return "Very Slightly Likely";
  }
  
  return "No Likelihood";

}

//======================================================================
//  Global functions to implement:

//      long       *        a  Goodness   yields a  Goodness
//      Goodness   * or /   a  long       yields a  Goodness

aipGoodness operator * (const aipGoodness&  x, 
			       const long y) {
  return aipGoodness ( x.val() * y );
}

aipGoodness operator * (const long  x, 
			       const aipGoodness&  y) {
  return aipGoodness ( x * y.val() );
}

aipGoodness operator / (const aipGoodness& x, 
			       const long y) {
  return aipGoodness ( x.val() / y );
}

aipGoodness& operator *= (aipGoodness& x, const long y) {
  x.set_val ( x.val() * y );
}

aipGoodness& operator /= (aipGoodness& x, const long y) {
  x.set_val ( x.val() / y );
}


//======================================================================
//  Global functions to implement:

//      Qualifier  applied to a  Goodness   yields a  Goodness
//      Goodness   applied to a  Qualifier  yields a  Goodness

aipGoodness operator *  (const aipGoodness& x,
			        const aipQualifier& y) {
  return aipGoodness(apply(x.val(), y.val()));
}

aipGoodness operator *  (const aipQualifier& x,
				const aipGoodness& y) {
  return aipGoodness(apply(y.val(), x.val()));
}

aipGoodness operator /  (const aipGoodness& x, 
			        const aipQualifier& y) {
  return aipGoodness(unapply(x.val(), y.val()));
}

aipGoodness& operator*= (aipGoodness& x, 
				 const aipQualifier& y) {
  x.set_val ( apply(x.val(), y.val()) );
  return x;
}

aipGoodness& operator /= (aipGoodness& x,
				 const aipQualifier& y) {
  x.set_val ( unapply(x.val(), y.val()) );
  return x;
}


//======================================================================
//  Global functions to implement:

//      Qualifier   applied to an  Importance  yields an  Importance
//      Importance  applied to a   Qualifier   yields an  Importance

aipImportance operator *  (const aipImportance& x,
				  const aipQualifier& y) {
  return aipImportance(apply(x.val(), y.val()));
}

aipImportance operator *  (const aipQualifier& x,
				  const aipImportance& y) {
  return aipImportance(apply(y.val(), x.val()));
}

aipImportance operator /  (const aipImportance& x,
				  const aipQualifier& y) {
  return aipImportance(unapply(x.val(), y.val()));
}

aipImportance& operator *= (aipImportance& x,  
				   const aipQualifier& y) {
  x.set_val(apply(x.val(), y.val()));
  return x;
}

aipImportance& operator /= (aipImportance& x,
				  const aipQualifier& y) {
  x.set_val(unapply(x.val(), y.val()));
  return x;
}


//======================================================================
//  Global functions to implement:

//      Likelihood  applied to a  Likelihood  yields a  Likelihood

aipLikelihood operator *  (const aipLikelihood& x,
				  const aipLikelihood& y) {
  return aipLikelihood(apply(x.val(), y.val()));
}

aipLikelihood operator /  (const aipLikelihood& x,
				  const aipLikelihood& y) {
  return aipLikelihood(unapply(x.val(), y.val()));
}

aipLikelihood& operator *= (aipLikelihood& x,
				  const aipLikelihood& y) {
  x.set_val(apply(x.val(), y.val()));
}

aipLikelihood& operator /= (aipLikelihood& x,
				  const aipLikelihood& y) {
  x.set_val(unapply(x.val(), y.val()));
}


//======================================================================
//                           License
//
//   Permission is hereby granted, free of charge, to any 
//   person obtaining a copy of this software and associated 
//   documentation files (the "Software"), to deal in the Software 
//   without restriction, including without limitation the rights 
//   to use, copy, modify, merge, publish, distribute, sublicense, 
//   and/or sell copies of the Software, and to permit persons to 
//   whom the Software is furnished to do so, subject to the 
//   following conditions:
//
//   The copyright notice and this license shall be included in all 
//   copies or substantial portions of the Software.
//
//   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
//   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES 
//   OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
//   NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
//   HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
//   WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
//   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
//   OTHER DEALINGS IN THE SOFTWARE.
//
//
//**********************************************************************

