//**********************************************************************
//  aipGood.h   -   Goodness, Importance, Likelihood
//
//  This header file...
//     -  Declares the following classes:
//           aipGoodness     How good or bad something is
//           aipImportance   How important something is
//           aipLikelihood   Likelihood of occurring or being true
//           in addition to various base classes.
//     - Defines constants of these types.
//     - Declares functions to operate on these types. 
//
//  Copyright Brian Marshall 1999
//
//  See the license at end of this file.
//
//  99/03/07  Brian Marshall  development begun.
//  99/05/01  Brian Marshall  naming conventions changed.
//  99/05/24  Brian Marshall  added  goodness * and / longs.
//
//----------------------------------------------------------------------

#ifndef aipGood_h_guard
#define aipGood_h_guard

//----------------------------------------------------------------------
//  Goodness, Importance, Likelihood
//
//  Goodness is a measure of how good or bad something is.
//
//  Goodnesses may be combined by adding.  The goodness of two things
//  taken together is the sum of their goodnesses.  The affect of a
//  goodness in a sum may be removed by subtracting it from the sum.
//
//  Importance and Likelihood may be applied to goodnesses 
//  and to each other by multiplying.  The effect of such a 
//  multiplication may be removed by dividing.
//
//        Being able to sleep until 7:00 am is 'Quite-Good'.  
//        Being able to catch the news at 7:15 is 'Good', 
//        but this is only 'Slightly-Important'.  If I set 
//        the alarm for 7:00, it is 'Fairly-Unlikely' that 
//        I will be late for work, which would be 'Bad' 
//        and 'Very-Important'.
//
//   aipGoodness g_sleep_to_7 = aipQuiteGood;
//   aipGoodness g_715_news = aipGood;
//   aipImportance i_715_news = aipSlightlyImportant;
//   g_715_news *= i_715_news;             // Apply the importance.
//   aipGoodness g_late_for_work = aipBad * aipVeryImportant;
//   g_late_for_work *= aipFairlyUnLikely;
//   aipGoodness g_up_at_7 = g_sleep_to_7 + g_715_news + g_late_for_work;
//   cout << "Getting up at seven is " 
//        << g_up_at_7.description() << "\n";
//
//        If we have made a similar evaluation for getting
//        up at 7:30, we can compare them:
//
//   if (g_up_at_700 > g_up_at_730) 
//       cout << "Get up at 7:00." << "\n";
//   else
//       cout << "Get up at 7:30." << "\n";
//
//----------------------------------------------------------------------
//  Combining goodnesses and qualifiers.
//
//
//  aipGoodness     +        aipGoodness     is a   aipGoodness
//
//  aipGoodness     * or /   aipImportance   is a   aipGoodness
//  aipGoodness     * or /   aipLikelihood   is a   aipGoodness
//
//  aipImportance   * or /   aipImportance   is a   aipImportance
//  aipImportance   * or /   aipLikelihood   is a   aipImportance
//
//  aipLikelihood   * or /   aipLikelihood   is a   aipLikelihood
//
//----------------------------------------------------------------------
//  Classes.  Sub-classing is shown by indentation.

class aipGBase;            // Base class.

  class  aipGoodness;      // How good or bad something is.

  class  aipQualifier;     // A quantity that qualifies a goodness.
    class  aipImportance;  // How important something is.
    class  aipLikelihood;  // How likely something is to occur.


//======================================================================
//  aipGBase   -   Abstract base class for goodness and qualifiers.
//
//  Note that the constructor taking a long is provided primarily
//  for the use of programmers defining constants for subclasses.
//
//  val() and set_val() functions are provided for the use of
//  programmers writing i/o functions.
//
//  This is an abstract class; concrete subclasses will override
//  the description() function.


class aipGBase {

protected:

    long  m_val;   // Internal representation available to subclasses.

public:

    aipGBase (void);
    aipGBase (const aipGBase& x);
    aipGBase (long lval);
    virtual ~aipGBase (void);

    virtual aipGBase& operator = (const aipGBase& x);

    bool operator <  (const aipGBase& x) const;
    bool operator >  (const aipGBase& x) const;
    bool operator == (const aipGBase& x) const;
    bool operator != (const aipGBase& x) const;
    bool operator <= (const aipGBase& x) const;
    bool operator >= (const aipGBase& x) const;

    long val (void) const;

    void set_val (const long x);

    virtual const char * description (void) const =0;

};


//======================================================================
//  aipGoodness  -  How good or bad something is.
//
//  aipGoodness objects may be assigned and compared.
//  aipGoodness objects may be combined by adding.


class aipGoodness : public aipGBase {
public:

    aipGoodness (void)                  : aipGBase()   {}
    aipGoodness (const aipGoodness& x)  : aipGBase(x)  {}
    aipGoodness (const long x)          : aipGBase(x)  {}

    virtual ~aipGoodness (void) {}

    aipGoodness operator +  (const aipGoodness& x) const;
    aipGoodness operator -  (const aipGoodness& x) const;

    aipGoodness& operator += (const aipGoodness& x);
    aipGoodness& operator -= (const aipGoodness& x);

    virtual const char * description (void) const;

};


//----------------------------------------------------------------------
//  aipGoodness constants on which the operation functions depend.

const aipGoodness aipForbidden         = aipGoodness( -99999 );
const aipGoodness aipVeryVeryBad       = aipGoodness(  -1024 );
const aipGoodness aipVeryBad           = aipGoodness(   -512 );
const aipGoodness aipQuiteBad          = aipGoodness(   -256 );
const aipGoodness aipBad               = aipGoodness(    -64 );
const aipGoodness aipFairlyBad         = aipGoodness(    -32 );
const aipGoodness aipSomewhatBad       = aipGoodness(     -8 );
const aipGoodness aipSlightlyBad       = aipGoodness(     -2 );
const aipGoodness aipVerySlightlyBad   = aipGoodness(     -1 );
const aipGoodness aipNeutral           = aipGoodness(      0 );
const aipGoodness aipVerySlightlyGood  = aipGoodness(      1 );
const aipGoodness aipSlightlyGood      = aipGoodness(      2 );
const aipGoodness aipSomewhatGood      = aipGoodness(      3 );
const aipGoodness aipFairlyGood        = aipGoodness(      4 );
const aipGoodness aipGood              = aipGoodness(      5 );
const aipGoodness aipQuiteGood         = aipGoodness(      6 );
const aipGoodness aipVeryGood          = aipGoodness(      7 );
const aipGoodness aipVeryVeryGood      = aipGoodness(      8 );
const aipGoodness aipManditory         = aipGoodness(  99999 );


//======================================================================
//  aipQualifier  -  Abstract base class for objects that qualify
//                  aipGoodnesses and each other.


class aipQualifier : public aipGBase {
public:

    aipQualifier (void)                  : aipGBase()   {}
    aipQualifier (const aipQualifier& x) : aipGBase(x)  {}
    aipQualifier (const long x)          : aipGBase(x)  {}

    virtual ~aipQualifier (void) {}

};

//======================================================================
//  aipImportance   -   How important something is.
//
//  aipImportance objects may be assigned and compared.
//
//  An aipImportance may modify an aipGoodness or aipLikelihood 
//  by multiplying.

class aipImportance : public aipQualifier {
public:

    aipImportance (void)                    : aipQualifier()   {}
    aipImportance (const aipImportance& x)  : aipQualifier(x)  {}
    aipImportance (const long x)            : aipQualifier(x)  {}

    virtual ~aipImportance (void) {}

    virtual const char * description (void) const;

};


//----------------------------------------------------------------------
//  aipImportance constants on which the operation functions depend.

const aipImportance aipNoImportance           = aipImportance(    0 );
const aipImportance aipVerySlightlyImportant  = aipImportance(    4 );
const aipImportance aipSlightlyImportant      = aipImportance(   16 );
const aipImportance aipSomewhatImportant      = aipImportance(   64 );
const aipImportance aipFairlyImportant        = aipImportance(  128 );
const aipImportance aipImportant              = aipImportance(  256 );
const aipImportance aipQuiteImportant         = aipImportance(  512 );
const aipImportance aipVeryImportant          = aipImportance( 1024 );
const aipImportance aipVeryVeryImportant      = aipImportance( 2048 );

const aipImportance aipFairlyUnImportant    = aipSomewhatImportant;
const aipImportance aipVeryUnImportant      = aipSlightlyImportant;
const aipImportance aipVeryVeryUnImportant  = aipVerySlightlyImportant;


//======================================================================
//  aipLikelihood   -   How likely something is to occur or be true.
//
//  aipLikelihood objects may be assigned and compared.
//
//  An aipImportance may modify an aipGoodness or aipLikelihood 
//  by multiplying.

class aipLikelihood : public aipQualifier {
public:

    aipLikelihood (void)                    : aipQualifier()   {}
    aipLikelihood (const aipLikelihood& x)  : aipQualifier(x)  {}
    aipLikelihood (const long x)            : aipQualifier(x)  {}

    virtual ~aipLikelihood (void) {}

    virtual const char * description (void) const;

};


//----------------------------------------------------------------------
//  aipLikelihood constants on which the operation functions depend.

const aipLikelihood aipNoLikelihood        = aipLikelihood(   0 );
const aipLikelihood aipVerySlightlyLikely  = aipLikelihood(   2 );
const aipLikelihood aipSlightlyLikely      = aipLikelihood(   8 );
const aipLikelihood aipSomewhatLikely      = aipLikelihood(  32 );
const aipLikelihood aipFairlyLikely        = aipLikelihood(  64 );
const aipLikelihood aipLikely              = aipLikelihood(  80 );
const aipLikelihood aipQuiteLikely         = aipLikelihood(  96 );
const aipLikelihood aipVeryLikely          = aipLikelihood( 112 );
const aipLikelihood aipVeryVeryLikely      = aipLikelihood( 126 );
const aipLikelihood aipCertain             = aipLikelihood( 128 );

const aipLikelihood aipFairlyUnLikely     = aipSomewhatLikely;
const aipLikelihood aipVeryUnLikely       = aipSlightlyLikely;
const aipLikelihood aipVeryVeryUnLikely   = aipVerySlightlyLikely;


//======================================================================
//  Global functions to implement:

//      Goodness   * or /   a  long       yields a  Goodness
//      long       *        a  Goodness   yields a  Goodness

aipGoodness operator * (const aipGoodness& x, const long y);
aipGoodness operator * (const long x, const aipGoodness& y);
aipGoodness operator / (const aipGoodness& x, const long y);

aipGoodness& operator *= (aipGoodness& x, const long y);
aipGoodness& operator /= (aipGoodness& x, const long y);


//======================================================================
//  Global functions to implement:

//      Qualifier  applied to a  Goodness   yields a  Goodness
//      Goodness   applied to a  Qualifier  yields a  Goodness

aipGoodness operator * (const aipGoodness&  x, const aipQualifier& y);
aipGoodness operator * (const aipQualifier& x, const aipGoodness&  y);
aipGoodness operator / (const aipGoodness&  x, const aipQualifier& y);

aipGoodness& operator*= (aipGoodness& x, const aipQualifier& y);
aipGoodness& operator /= (aipGoodness& x, const aipQualifier& y);


//======================================================================
//  Global functions to implement:

//      Qualifier   applied to an  Importance  yields an  Importance
//      Importance  applied to a   Qualifier   yields an  Importance

aipImportance operator *  (const aipImportance& x,
			    const aipQualifier&);

aipImportance operator *  (const aipQualifier& x,
                            const aipImportance&);

aipImportance operator /  (const aipImportance& x,
                            const aipQualifier&);

aipImportance& operator *= (aipImportance& x,
                            const aipQualifier& x);

aipImportance& operator /= (aipImportance& x,
                            const aipQualifier& x);


//======================================================================
//  Global functions to implement:

//      Likelihood  applied to a  Likelihood  yields a  Likelihood

aipLikelihood operator *  (const aipLikelihood& x,
                            const aipLikelihood& x);

aipLikelihood operator /  (const aipLikelihood& x,
                            const aipLikelihood& x);

aipLikelihood& operator *= (aipLikelihood& x,
                            const aipLikelihood& x);

aipLikelihood& operator /= (aipLikelihood& x,
                            const aipLikelihood& x);


//======================================================================

#endif

//----------------------------------------------------------------------
//                           License
//
//   Permission is hereby granted, free of charge, to any 
//   person obtaining a copy of this software and associated 
//   documentation files (the "Software"), to deal in the Software 
//   without restriction, including without limitation the rights 
//   to use, copy, modify, merge, publish, distribute, sublicense, 
//   and/or sell copies of the Software, and to permit persons to 
//   whom the Software is furnished to do so, subject to the 
//   following conditions:
//
//   The copyright notice and this license shall be included in all 
//   copies or substantial portions of the Software.
//
//   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
//   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES 
//   OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
//   NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
//   HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
//   WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
//   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
//   OTHER DEALINGS IN THE SOFTWARE.
//
//**********************************************************************

