//**********************************************************************
//  apprxVal.h   -   Approximate Values
//
//  This header file declares...
//    - Approximate values   which can be...
//      - values based on a sequence of noted values  which can be...
//        - various classes that approach this in different ways
//    - Components used in approximate value classes, like...
//      - mean and running-mean
//      - list of values
//      - histogram of values (array of bin-counts)
//      - meanogram (array of bin-counts and bin-sums)
//
//  Objects declared in this file throw aipXcpt exceptions on problems.
//
//  Copyright Brian Marshall 1999
//
//  See the license at end of this file.
//
//  98/07/04  Brian Marshall  development begun.
//  99/05/01  Brian Marshall  brought into AIParts project.
//
//----------------------------------------------------------------------

#ifndef apprxVal_h_guard
#define apprxVal_h_guard

#include "aipRge.h"
#include "aipXcpt.h"

//======================================================================
//  Constants.

const long Max_Neg_Long = -2000000000;
const long Max_Pos_Long =  2000000000;

//----------------------------------------------------------------------
//  Classes.  Subclassing is shown by indentation.

class apprxVal;           // Approximate value.
  class nvStat;           // Noted value statistic.
    class meanVal;        // Mean of accepted values.
      class nvsRunMean;   // Running mean of accepted values.
    class estVal;         // Estimated value.
    class pgAve;          // Power group average.
    class lvhStat;        // Limited-value histogram statistics.

    // Components available for use in nvStat classes...

class nvsMean;            // Mean - sum and count of values.
class nvsMeanSupl;        // Mean supplement.
class nvsValues;          // List of values.
class nvsHist;            // Histogram.
class nvsMeanogram;       // Meanogram - histogram of means.

//----------------------------------------------------------------------
//  A note on constructors, setup(), init() and reset().
//
//  Constructors initialize things that don't change even if reset
//  is called.  They set pointers where memory is to be newwed.
//
//  setup() is used to allocate memory and other tasks that are done
//  once getting an object ready to be used.  Functions specifying
//  various things can be called before setup() is called.  setup()
//  is called automatically before anything is done with an object.
//
//  init() is a private function in each class that initializes 
//  members.  It is called by the constructor (ie. before setup()
//  is called) and by reset().
//
//  reset() is a public function to make an object call init() on
//  itself, its members and its parent classes.
//
//======================================================================
//  apprxValBase   -   Base class for apprxVal and components.

class apprxValBase {

    bool   m_is_setup;

protected:

    virtual void setup (void) {
        if (m_is_setup) {
            throw aipXcpt ("apprxSetupup", 
                           "apprxValBase setup() twice");
        }
        m_is_setup = true;
    }

public:

    apprxValBase (void) { 
        m_is_setup = false;
    }

    virtual ~apprxValBase (void) {}

    bool  is_setup (void) const  { return m_is_setup; }

};

//======================================================================
//  apprxVal   -   Approximate Value
//
//  An apprxVal object has a current value and it can be asked whether
//  it is approximately equal to a specified long value.
//
//  This class will presumably be converted to a template eventually.
//  In the mean time, at least, it can be used by people with C++
//  compliers that don't support templates well.


class apprxVal : public apprxValBase {

    long   m_val;               // Current idea of value.

    bool   m_val_is_set;
    bool   m_requires_calc;

    long   m_max_pos_dev;
    long   m_max_neg_dev;
    bool   m_auto_dev;

    void init (void) {
        m_val = 0;
        m_val_is_set = false;
        m_requires_calc = false;
    }

protected:

    virtual void setup (void) {
        apprxValBase::setup();
    }

    virtual void calculate (void) {  // Override and call this.
        m_requires_calc = false;
    }

    virtual void determine_deviation (void) {
        if (m_val_is_set) {
            set_max_deviation ( val() / 4 + 1 );
        } else {
            set_max_deviation ( 1 );
        }
    }

    virtual void on_change (void) {
        determine_deviation();
    }

public:

    apprxVal (void) { 
        m_max_pos_dev = +1;
        m_max_neg_dev = -1;
        m_auto_dev = true;
        init(); 
    }
    virtual ~apprxVal (void) {}

    virtual void reset (void) {
        if ( ! is_setup() ) setup();
        init();
    }

    virtual void force (long x) {
        reset();
        set_val(x);
    }

    long  val (void) {
        if ( ! is_setup() ) setup();
        if ( m_requires_calc ) calculate();
        return m_val; 
    }

    long operator * (void)  { return val(); }

    long latest_val (void) const {
        return m_val;
    }

    bool  val_is_set    (void) const  { return m_val_is_set;    }
    bool  requires_calc (void) const  { return m_requires_calc; }
    long  max_pos_dev   (void) const  { return m_max_pos_dev;   }
    long  max_neg_dev   (void) const  { return m_max_neg_dev;   }

    void set_val (long x) {
        if ( ! is_setup() ) setup();
        if (x != m_val || ! m_val_is_set) {
            m_val = x; 
            m_val_is_set = true;
            on_change();
        }
    }

    void set_max_pos_dev  (long x)  { 
        m_max_pos_dev  = x; 
        m_auto_dev = false;
    }

    void set_max_neg_dev  (long x)  { 
        m_max_neg_dev  = x; 
        m_auto_dev = false;
    }

    void set_auto_dev     (bool x)  { m_auto_dev     = x; }

    void set_requires_calc (void) { m_requires_calc = true; }

    void set_max_deviation (long x) { 
        m_max_pos_dev = x;
        m_max_neg_dev = 0 - x; 
        m_auto_dev = false;
    }

    bool approx_equal (long x) {
        long dev = x - val();
        return (dev >= m_max_neg_dev && dev <= m_max_pos_dev);
    }

};


//----------------------------------------------------------------------
//  apprxFVal   -   Approximate Value with a float value.
//
//  An apprxVal object has a current value and it can be asked whether
//  it is approximately equal to a specified long value.
//
//  Yeah, I'm embarassed about this - the apprxVal class should be
//  a template so people can use longs or floats or whatever.  I
//  will convert this stuff to templates eventually.  One advantage
//  of the way things are now is that it works even if you have a
//  C++ compiler that doesn't support templates well.

class apprxFVal : public apprxVal {

    float  m_float_val;

public:

    apprxFVal (void) { 
        m_float_val = 0.0;
    }
    virtual ~apprxFVal (void) {}

    float float_val (void) { return m_float_val; }

    void set_float_val (float x) {
        m_float_val = x;
        apprxVal::set_val ( long(x+0.5) );
    }

};


//----------------------------------------------------------------------
//  Global functions to determine if a long value is approximately
//  equal to an apprxVal value or two apprxVal values are approximately
//  equal.

bool operator == (const long x, apprxVal& apv);

bool operator == (apprxVal& apv, const long x);

bool operator == (apprxVal& apv1, apprxVal& apv2);


//======================================================================
//  nvStat   -   Noted Value Statistic
//
//  An nvStat object is an approximate-value which is based on values
//  which have been noted.
//
//  An nvStat object is given a number of values to note.  At any
//  time after the first value, the object may be asked for:
//      - the 'true' value
//      - whether a value is approximately equal to the true value
//
//  Subclasses will override note_val().  The note_val() provided
//  by this class makes the current value the last value noted.
//
//  If subclasses do not set the value on every call to note(),
//  they should override calculate().  This function is called by
//  val() and the * operator if m_val_is_set is false.
//
//  m_max_pos_dev and m_max_neg_dev define the most positive and most
//  negative the difference between a number and m_val can be for the 
//  number to be considered to be approximately equal to m_val.

class nvStat : public apprxVal {

    long   m_min;
    long   m_max;
    bool   m_limit_min_max;

    long   m_num_noted;
    long   m_max_num_noted;

    long   m_num_accepted;

    void init (void) {
        m_max = Max_Neg_Long;
        m_min = Max_Pos_Long;
        m_max_num_noted = 10000;
        m_num_noted = 0;
        m_num_accepted = 0;
        m_limit_min_max = false;
    }

protected:

    virtual void setup (void) {
        apprxVal::setup();
    }

    long  max_num_noted  (void) const  { return m_max_num_noted;  }
    bool  limit_min_max  (void) const  { return m_limit_min_max;  }

    void set_min           (long x)  { m_min           = x; }
    void set_max           (long x)  { m_max           = x; }
    void set_max_num_noted (long x)  { m_max_num_noted = x; }
    void set_limit_min_max (bool x)  { m_limit_min_max = x; }

    virtual void note_val (long x) {    // Override and call this.
        set_requires_calc();
    }

    virtual void calculate (void) {     // Override and call this.
        apprxVal::calculate();
    }

    virtual bool accept_val (long x) {  // Maybe override this.
        return true;
    }

public:

    nvStat (void) { init(); }
    virtual ~nvStat (void) {}

    virtual void reset (void) {
        apprxVal::reset();
        init();
    }

    virtual void force (long x) {
        reset();
        for (int i=0; i<8; i++) note(x);
    }

    long  min          (void) const  { return m_min;          }
    long  max          (void) const  { return m_max;          }
    long  num_noted    (void) const  { return m_num_noted;    }
    long  num_accepted (void) const  { return m_num_accepted; }

    void note (long x) {
        if ( ! is_setup() ) setup();
        if (m_num_noted >= m_max_num_noted) return;
        m_num_noted++;
        if (m_limit_min_max) {    // min and max are prescriptive.
            if (x < m_min) return;
            if (x > m_max) return;
        }
        if ( ! accept_val(x) ) return;
        m_num_accepted++;
        note_val(x);
        if (!m_limit_min_max) {   // min and max are descriptive.
            if (x < m_min) m_min = x;
            if (x > m_max) m_max = x;
        }
    }

};


//======================================================================
//  Components used in sub-classes of nvStat
//
//----------------------------------------------------------------------
//  nvsMean   -  Mean.

class nvsMean : public apprxValBase {

    long    m_sum;
    long    m_count;

    void init (void) {
        m_sum = 0;
        m_count = 0;
    }

public:

    virtual void setup (void) {
        apprxValBase::setup();
    }

    nvsMean (void) { init(); }
    virtual ~nvsMean (void) {}

    virtual void reset (void) { init(); }

    long sum   (void)  { return m_sum;   }
    long count (void)  { return m_count; }
        
    void set_sum   (long x)  { m_sum = x;   }
    void set_count (long x)  { m_count = x; }

    void sum_plus   (long x) { m_sum += x;   }
    void count_plus (long x) { m_count += x; }

    void sum_minus   (long x) { m_sum -= x;   }
    void count_minus (long x) { m_count -= x; }

    long val (void) {
        if (m_count > 0) {
            return m_sum / m_count;
        }
        return 0;
    }

    void note (long x) {
        m_sum += x;
        m_count++;
    }

};

//----------------------------------------------------------------------
//  nvsRunMean   -  Running Mean.

class nvsRunMean : public nvsMean {

    long   * m_circ;         // Array as a circular list.
    long     m_run_length;   // Num vals saved; m_circ is 1 larger.
    long     m_next;         // Element in m_circ to get next val.

    void init (void) {
        m_next = 0;
    }

public:

    virtual void setup (void) {
        nvsMean::setup();
        if (m_run_length < 1 || m_run_length > 1000)
            throw aipXcpt ("nvsRMlen", 
                           "nvsRunMean run len bad");
        m_circ = new long [m_run_length+1];
        if ( ! m_circ ) throw aipXcpt ("nvsRMmem", 
                                       "nvsRunMean mem fail");
    }

    nvsRunMean (void) {
        m_circ = 0;
        m_run_length = 16;
        init();
    }

    virtual ~nvsRunMean (void) {
        if (m_circ) delete [] m_circ;
    }

    virtual void reset (void) {
        nvsMean::reset();
        init();
    }
        
    long run_length (void)  { return m_run_length; }

    void set_run_length (long x) { m_run_length = x; }

    void note (long x) {
        m_circ[m_next] = x;
        if ( ++m_next > m_run_length ) m_next = 0;
        sum_plus (x);
        if ( count() < m_run_length ) {
           count_plus(1);
        } else {
           sum_minus ( m_circ[m_next] );
        }
    }

};

//----------------------------------------------------------------------
//  nvsMeanSupl   -   Mean Supplement
//
//  A mean-supplement is more sum and count that might be included in 
//  a mean.

class nvsMeanSupl : public apprxValBase {

    nvsMean  m_mean;         // Sum and count of noted values.
    bool     m_should;       // This should be applied.
    bool     m_applied;      // This has been applied.

    aipRge   m_rge_accept;   // Range of values accepted for noting.

    void init (void) {
        m_mean.reset();
        m_should = m_applied = false;
    }

public:

    virtual void setup (void) {
        apprxValBase::setup();
    }

    nvsMeanSupl (void) { init(); }
    virtual ~nvsMeanSupl (void) {}

    virtual void reset (void) { init(); }

    long     sum        (void)  { return m_mean.sum();   }
    long     count      (void)  { return m_mean.count(); }
    bool     should     (void)  { return m_should;       }
    bool     applied    (void)  { return m_applied;      }
    aipRge&  rge_accept (void)  { return m_rge_accept;   }

    void note (long x) { 
        if ( ! m_rge_accept.inverted() &&
             ! m_rge_accept.contains(x) ) return;
        m_mean.note(x); 
    }

    void set_should  (bool should  =true)  { m_should  = should;  }
    void set_applied (bool applied =true)  { m_applied = applied; }

    void set_rge_accept (aipRge& rge) {
        m_rge_accept.set(rge);
    }

    void set_rge_accept (long min, long max) {
        m_rge_accept.set(min,max);
    }

};


//----------------------------------------------------------------------
//  Global functions for adding and removing mean-supplements.

nvsMean& operator += (nvsMean& mean, nvsMeanSupl& supl);

nvsMean& operator -= (nvsMean& mean, nvsMeanSupl& supl);

//----------------------------------------------------------------------
//  nvsValues   -    A set of values.

class nvsValues : public apprxValBase {

    long      m_num_val;
    long      m_max_num_val;
    long    * m_val;           // Array of values.

    aipRge    m_rge;           // Range of values.
    nvsMean   m_mean;          // Sum and count of values.

    void init (void) {
        m_num_val = 0;
        m_rge.set (Max_Pos_Long, Max_Neg_Long);
        m_mean.reset();
    }

public:

    virtual void setup (void) {
        apprxValBase::setup();
        if (m_max_num_val < 1 || m_max_num_val > 100000) 
            throw aipXcpt ("vnsValNum","nvsValues: bad max_num_val");
        m_val = new long[m_max_num_val];
        if (!m_val) throw aipXcpt ("nvsValMem","nvsValues: mem fail");
    }

    nvsValues (long max_num_val =0) { 
        m_val = 0;
        m_max_num_val = max_num_val; 
        init();
    }

    virtual ~nvsValues (void) {
        if (m_val) delete [] m_val;
    }

    virtual void reset (void) { init(); }

    void set_max_num_val (long x) { m_max_num_val = x; }

    long max_num_val  (void) { return m_max_num_val;  } // 0 = bad.
    long num_val      (void) { return m_num_val;      }
    long min          (void) { return m_rge.min();    }
    long max          (void) { return m_rge.max();    }
    long sum          (void) { return m_mean.sum();   }
    long count        (void) { return m_mean.count(); }
    long mean         (void) { return m_mean.val();   }

    void note (long x) {
        if (m_num_val >= m_max_num_val) return;
        m_val[m_num_val++] = x;
        m_rge.note_included(x);
        m_mean.note(x);
    }

    long val (long ival) {
        if (ival < 0) ival = 0;
        if (ival >= m_num_val) ival = m_num_val-1;
        return m_val[ival];
    }

    void pluseq (long x) {
        for (long i=0; i<m_num_val; i++) {
            m_val[i] += x;
        }
    }

    void minuseq (long x) {
        for (long i=0; i<m_num_val; i++) {
            m_val[i] -= x;
        }
    }

    void multiplyeq (long x) {
        for (long i=0; i<m_num_val; i++) {
            m_val[i] /= x;
        }
    }

    void divideeq (long x) {
        for (long i=0; i<m_num_val; i++) {
            m_val[i] *= x;
        }
    }

    nvsValues& operator += (long x) { pluseq(x);      return *this; }
    nvsValues& operator -= (long x) { minuseq(x);     return *this; }
    nvsValues& operator *= (long x) { multiplyeq(x);  return *this; }
    nvsValues& operator /= (long x) { divideeq(x);    return *this; }

    long limited_sum (long min, long max) {
        long the_sum = 0;
        for (long i=0; i<m_num_val; i++) {
            if (m_val[i] >= min && m_val[i] <= max) {
                the_sum += m_val[i];
            }
        }
        return the_sum;
    }

    long limited_count (long min, long max) {
        long the_count = 0;
        for (long i=0; i<m_num_val; i++) {
            if (m_val[i] >= min && m_val[i] <= max) {
                the_count++;
            }
        }
        return the_count;
    }

};


//----------------------------------------------------------------------
//  nvsHist   -   Histogram - array of counts.
//
//  ibin = (val - m_offset) / m_bin_size;
//
//  Note that a triple is a set of three consecutive bins and a
//  triple is referred to as the index of the center bin.

class nvsHist : apprxValBase {

    long    m_num_bin;
    long  * m_count;            // Array of counts.
    long    m_last_noted_bin;
    long    m_offset;
    long    m_bin_size;
    bool    m_rge_except;       // true: exception on out or range.

    void init (void) {
        m_last_noted_bin = 0;
        m_offset = 0;                  // Default.
        m_bin_size = 1;                // Default.
    }

public:

    virtual void setup (void) {
        apprxValBase::setup();
        if (m_num_bin < 1 || m_num_bin > 1000000)
            throw aipXcpt ("nvsHistNum","nvsHist bad num-bin");
        m_count = new long[m_num_bin];
        if (!m_count) throw aipXcpt("nvsHistMem", "nvsHist mem fail");
        for (long i=0; i<m_num_bin; i++) {
            m_count[i] = 0;
        }
    }

    nvsHist (long num_bin =0) {
        m_count = 0;
        m_num_bin = num_bin;
        m_rge_except = false;
        init();
    }

    virtual ~nvsHist (void) {
        if (m_count) delete [] m_count;
    }

    virtual void reset (void) {
        for (long i=0; i<m_num_bin; i++) {
            m_count[i] = 0;
        }
        init();
    }

    long num_bin        (void) const { return m_num_bin;        }
    long last_noted_bin (void) const { return m_last_noted_bin; }
    long offset         (void) const { return m_offset;         }
    long bin_size       (void) const { return m_bin_size;       }
    bool rge_except     (void) const { return m_rge_except;     }

    void set_num_bin    (long x)  { m_num_bin = x;    }
    void set_offset     (long x)  { m_offset = x;     }
    void set_bin_size   (long x)  { m_bin_size = x;   }
    void set_rge_except (bool x)  { m_rge_except = x; }

    void configure (long min, long max) {
        if (m_num_bin < 1 || m_num_bin > 1000000)
            throw aipXcpt ("nvsHistNum2",
                           "nvsHist configure bad num-bin");
        set_offset (min);
        m_bin_size = (max - min) / m_num_bin;
        if (m_bin_size < 1) m_bin_size = 1;
        if ( (max-min)/m_bin_size >= m_num_bin) m_bin_size++;
    }

    void note (long x) {
        long ibin = (x - m_offset) / m_bin_size;
        if (ibin < 0 || ibin >= m_num_bin) {
            m_last_noted_bin = -1;
            if (m_rge_except) {
                throw aipXcpt ("nvsHistRge",
                               "nvsHist val out-of-range");
            } else {
                return;
            }
        }
        (m_count[ibin])++;
        m_last_noted_bin = ibin;
    }

    void note (nvsValues& values) {
        for (int ival=0; ival<values.num_val(); ival++) {
            note (values.val(ival) );
        }
    }

    long count (long ibin) {
        if (ibin < 0) ibin = 0;
        if (ibin >= m_num_bin) ibin = m_num_bin-1;
        return m_count[ibin];
    }

    long biggest_bin_index (void) {
        long ibig = 0;
        long biggest = 0;
        for (long ibin=0; ibin<m_num_bin; ibin++) {
            if (m_count[ibin] > biggest) {
                ibig = ibin;
                biggest = m_count[ibin];
            }
        }
        return ibig;
    }

    long biggest_triple (void) const {
        if (m_num_bin < 3) return -1;
        long ibig = 1;
        long biggest = 0;
        for (long ibin=1; ibin<m_num_bin-1; ibin++) {
            long count3 = m_count[ibin-1] + 
                          m_count[ibin]   + 
                          m_count[ibin+1];
            if (count3 > biggest) {
                ibig = ibin;
                biggest = count3;
            }
        }
        return ibig;
    }

    long biggest_bin_count (void) {
        long ibig = biggest_bin_index();
        if (ibig < 0) return 0;
        return m_count[ibig];
    }

    long triple_count (long itriple) {
        if (itriple < 0 || itriple > m_num_bin) return 0;
        long count3 = m_count[itriple];
        if (itriple-1 >= 0)           count3 += m_count[itriple-1];
        if (itriple+1 <= num_bin()-1) count3 += m_count[itriple+1];
        return count3;
    }

};


//----------------------------------------------------------------------
//  nvsMeanogram   -   Array of bins of counts and sums.
//
//  A nvsMeanogram is like a histogram except that, in addition to
//  having a count in each bin, there is also a sum.

class nvsMeanogram : public nvsHist {

    long  * m_sum;        // Array of sums.

    void init (void) {
    }

public:

    virtual void setup (void) {
        nvsHist::setup();
        if (num_bin() < 1 || num_bin() > 100000) 
            throw aipXcpt ("nvsMeanNum","nvsMeanogram bad num-bin");
        m_sum = new long[num_bin()];
        if (!m_sum) throw aipXcpt("nvsMeano1", 
                                  "nvsMeanogram: mem fail: m_sum");
        for (long i=0; i<num_bin(); i++) {
            m_sum[i] = 0;
        }
    }

    nvsMeanogram (long num_bin =100) : nvsHist(num_bin) { 
        m_sum = 0; 
        init();
    }

    virtual ~nvsMeanogram (void) {
        if (m_sum)   delete [] m_sum;
    }

    virtual void reset (void) {
        nvsHist::reset();
        for (long i=0; i<num_bin(); i++) {
            m_sum[i] = 0;
        }
        init();
    }

    void note (long x) {
        nvsHist::note(x);
        if (last_noted_bin() >= 0) {
            m_sum[last_noted_bin()] += x;
        }
    }

    void note (nvsValues& values) {
        for (int ival=0; ival<values.num_val(); ival++) {
            note (values.val(ival) );
        }
    }

    long sum (long ibin) {
        if (ibin < 0) ibin = 0;
        if (ibin >= num_bin()) ibin = num_bin()-1;
        return m_sum[ibin];
    }

    long biggest_bin_sum (void) {
        long ibig = biggest_bin_index();
        if (ibig < 0) return 0;
        return m_sum[ibig];
    }

    long triple_sum (long itriple) {
        if (itriple < 0 || itriple > num_bin()) return 0;
        long sum3 = m_sum[itriple];
        if (itriple-1 >= 0)           sum3 += m_sum[itriple-1];
        if (itriple+1 <= num_bin()-1) sum3 += m_sum[itriple+1];
        return sum3;
    }


};


//======================================================================
//  Subclasses of nvStat
//
//----------------------------------------------------------------------
//  meanVal   -   Mean Value of all accepted values

class meanVal : public nvStat {

    nvsMean  m_mean;

    void init (void) {
        nvStat::reset();
        m_mean.reset();
    }

protected:

    virtual void setup (void) {
        nvStat::setup();
        m_mean.setup();
    }

    virtual void calculate (void) {
        set_val ( m_mean.val() );
        nvStat::calculate();
    }

    virtual void note_val (long x) {
        m_mean.note(x);
        nvStat::note_val(x);
    }

    void sum_plus   (long x)  { m_mean.sum_plus(x);   }
    void count_plus (long x)  { m_mean.count_plus(x); }

    void sum_minus   (long x)  { m_mean.sum_minus(x);   }
    void count_minus (long x)  { m_mean.count_minus(x); }

    void add_supl (nvsMeanSupl& supl) {
        sum_plus   ( supl.sum()   );
        count_plus ( supl.count() );
        supl.set_applied(true);
    }

    void remove_supl (nvsMeanSupl& supl) {
        sum_minus   ( supl.sum()   );
        count_minus ( supl.count() );
        supl.set_applied(false);
    }

public:

    meanVal (void) { init(); }
    virtual ~meanVal (void) {}

    virtual void reset (void) { 
        nvStat::reset();
        init(); 
    }

};

//----------------------------------------------------------------------
//  apprxMean   -   Mean of values approx equal to current value.
//
//  An apprxVal is a meanVal in which only values approximately
//  equal to the current value are accepted.

class apprxMean : public meanVal {

    void init (void) {
    }

protected:

    virtual bool accept_val (long x) {
        if ( ! val_is_set() ) return true;
        return approx_equal(x);
    }

public:

    apprxMean (void) { init(); }
    virtual ~apprxMean (void) {}

    virtual void reset (void) { 
        meanVal::reset();
        init(); 
    }

};


//----------------------------------------------------------------------
//  runVal   -   Running Mean Value of all accepted values

class runVal : public nvStat {

    nvsRunMean  m_run;

    void init (void) {
        m_run.reset();
    }

protected:

    virtual void setup (void) {
        nvStat::setup();
        m_run.setup();
    }

    virtual void calculate (void) {
        set_val ( m_run.val() );
        nvStat::calculate();
    }

    virtual void note_val (long x) {
        m_run.note(x);
        nvStat::note_val(x);
    }

public:

    runVal (void) { init(); }
    virtual ~runVal (void) {}

    virtual void reset (void) {
        nvStat::reset();
        init();
    }

    void set_run_length (long x) { m_run.set_run_length(x); }

};

//----------------------------------------------------------------------
//  apprxRun   -   Running Mean of vals approx equal to current value.
//
//  An approxRun is a runVal in which only values approximately
//  equal to the current value are accepted.

class apprxRun : public runVal {

    void init (void) {
    }

protected:

    virtual bool accept_val (long x) {
        if ( ! val_is_set() ) return true;
        return approx_equal(x);
    }

public:

    apprxRun (void) { init(); }
    virtual ~apprxRun (void) {}

    virtual void reset (void) {
        runVal::reset();
        init();
    }

};


//----------------------------------------------------------------------
//  pgAve   -   Power Group Average
//
//  'You only get counted if you are part of the biggest group.'
//  Noted values are dropped into a meanogram - a histogram where each
//  bin holds a count and a sum.  The value of a pgAve is the mean of
//  the values in the set of three consecutive bins with the most
//  values.
//
//  The meanogram is setup after m_pre_samp values have been accepted.
//  The range and bin-size of the meanogram is determined by the first
//  m_pre_samp accepted values.  Before this point, values are stored,
//  and val() returns the mean of these values other than the largest
//  and smallest.
//
//  Used to identify the characteristic value in a set that includes
//  a relatively high number of values that are correct plus or minus
//  some error, in addition to a variety of other values.

class pgAve : public nvStat {

    bool            m_accumulating_values;

    nvsValues       m_values;     // List of initial values.

    nvsMeanogram    m_meano;      // Meanogram of accepted values.

    float           m_float_val;  // Value calculated as a float.

    void init (void) {
        m_accumulating_values = true;
    }

protected:

    virtual void setup (void) {
        nvStat::setup();
        m_values.setup();
        m_meano.setup();
    }

    virtual void note_val (long x) {
        if (m_accumulating_values) {
            m_values.note(x);
            if (m_values.num_val() >= m_values.max_num_val()) {
                stop_accumulating_values();
            }
        } else {
            m_meano.note(x);
        }
        nvStat::note_val(x);
    }

    virtual void calculate (void) {
        long the_sum, the_count;
        if (m_accumulating_values) {
            the_sum = m_values.sum();
            the_count = m_values.count();
            if (m_values.num_val() > 2) {
                the_sum -= (m_values.min() + m_values.max());
                the_count -= 2;
            }
        } else {
            long itriple = m_meano.biggest_triple();
            the_sum = m_meano.triple_sum(itriple);
            the_count = m_meano.triple_count(itriple);
        }
        if (the_count > 0) {
            m_float_val = float(the_sum) / float(the_count);
            set_val ( long(m_float_val + 0.5) );
        }
        nvStat::calculate();
    }

public:

    pgAve (long max_init_val =10, long num_bin =16) 
            : m_values(max_init_val), m_meano(num_bin) {
        init();
    }

    virtual ~pgAve (void) {}

    virtual void reset (void) { 
        nvStat::reset(); 
        m_values.reset();
        m_meano.reset();
        init();
    }

    float  float_val (void) { 
        if ( requires_calc() ) calculate();
        return m_float_val; 
    }

    void set_max_init_val (long x) { m_values.set_max_num_val(x); }
    void set_num_bin      (long x) { m_meano.set_num_bin(x);      }

    bool approx_equal (long x) {
        if (m_accumulating_values) {
            return nvStat::approx_equal(x);
        } 
        long ibin = (x - m_meano.offset()) / m_meano.bin_size();
        if ( ibin < -1 || ibin > m_meano.num_bin() ) return false;
        long itriple = m_meano.biggest_triple();
        long dif = itriple - ibin;
        if (dif < -1 || dif > 1) return false;
        return true;
    }

    void stop_accumulating_values (void) {
        if (!m_accumulating_values) return;
        m_accumulating_values = false;
        m_meano.configure (min(), max());
        m_meano.note (m_values);
        set_limit_min_max(true);
    }

};

//----------------------------------------------------------------------
//  Limited-Value-Histogram Statistics.
//
//  Statistics based on histogram with a bin for each value in
//  a range.  This class descends from nvMeanStat, so val() returns
//  the mean of vaues in the range.

class lvhStat : public meanVal {

    nvsHist   * m_hist;

    void init (void) {
    }

protected:

    virtual void setup (void) {
        meanVal::setup();
        long num_bin = (max() - min()) + 1;
        if (num_bin < 1 || num_bin > 100000) 
            throw aipXcpt("lvhStat1", "lvhStat: bad num bin");
        m_hist = new nvsHist(num_bin);
        if ( ! m_hist ) 
            throw aipXcpt("lvhStat2", "lvhStat: mem fail: m_hist");
        m_hist->setup();
        m_hist->reset();
    }

    void note_val (long x) { 
        if (m_hist) m_hist->note(x-min());
        meanVal::note_val(x);
    }

public:

    lvhStat (long the_min, long the_max) {
        m_hist = 0;
        set_min(the_min);
        set_max(the_max);
        set_limit_min_max(true);
    }

    virtual ~lvhStat (void) {
        if (m_hist) delete m_hist;
    }

    virtual void reset (void) {
        meanVal::reset();
        m_hist->reset();
    }

    long highest_common_val (long min_bin =0) {
        long the_biggest_bin = m_hist->biggest_bin_count();
        long min_acceptable = the_biggest_bin / 2;
        int ibin = m_hist->num_bin() - 1;
        for ( ; ibin>=0; ibin--) {
            if (m_hist->count(ibin) > min_acceptable) {
                return min()+ibin;
            }
        }
        return min();
    }

    long most_common_val (void) {
        return m_hist->biggest_bin_count();
    }

};


//======================================================================

#endif

//----------------------------------------------------------------------
//                           License
//
//   Permission is hereby granted, free of charge, to any 
//   person obtaining a copy of this software and associated 
//   documentation files (the "Software"), to deal in the Software 
//   without restriction, including without limitation the rights 
//   to use, copy, modify, merge, publish, distribute, sublicense, 
//   and/or sell copies of the Software, and to permit persons to 
//   whom the Software is furnished to do so, subject to the 
//   following conditions:
//
//   The copyright notice and this license shall be included in all 
//   copies or substantial portions of the Software.
//
//   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
//   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES 
//   OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
//   NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
//   HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
//   WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
//   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
//   OTHER DEALINGS IN THE SOFTWARE.
//
//
//**********************************************************************

