//**********************************************************************
//  dirthard.cxx   -   Member functions for dirthard.h
//
//  Copyright Brian Marshall 1999
//
//  See the license at end of this file.
//
//  Brian Marshall    bmarshal@agt.net    (403)651-0584
//
//  99/05/27  Brian Marshall  development begun.
//  99/06/23  Brian Marshall  rework note() and event-passing.
//
//----------------------------------------------------------------------

#include "dirthard.h"

//----------------------------------------------------------------------
//  Notes.
//
//  If this file is compiled with  dhNoChk  defined, checking will
//  not be done to ensure that valid pointers are returned by 
//  functions.
//
//
//======================================================================
//  dhComponent   -   Component of a dirty-hard problem tree.
//
//----------------------------------------------------------------------
//  Construct a component.

dhComponent::dhComponent (void) {

  m_parent = 0;

  m_problem_aspect = m_strategy_aspect = 0;

}


//----------------------------------------------------------------------
//  Destruct a component.

dhComponent::~dhComponent (void) {

  if (m_problem_aspect)  delete m_problem_aspect;
  if (m_strategy_aspect) delete m_strategy_aspect;

}


//----------------------------------------------------------------------
//  Note that an event has occurred.
//
//  An event is 'bubbling' if both parents and children are to be
//  told and the event is being passed up to the root of the problem
//  tree.  Once it reaches the root, parents are no longer told and
//  the event is told recursively to all children of the root.


void dhComponent::note (dhEvent *x) {

  if ( x->bubbling() ) {
    if ( parent() ) {
      parent()->note(x);
      return;
    } else {                       // At top of problem tree.
      x->set_tell_parent(false);   // Not bubbling anymore.
    }
  }

  if (m_problem_aspect)  m_problem_aspect->note(x);
  if (m_strategy_aspect) m_strategy_aspect->note(x);
    
}


//----------------------------------------------------------------------
//  Evaluate the passed way.
//
//  If this component has a problem-domain specific aspect, let it
//  evaluate this way; otherwise, return aipNeutral.

aipGoodness dhComponent::evaluate (dhWay *way) {

  if (m_problem_aspect)  return m_problem_aspect->evaluate(way);

  return aipNeutral;

}


//======================================================================
//  dhOption   -   Possibility in a decision.
//
//----------------------------------------------------------------------
//  Construct an option.

dhOption::dhOption (void) {

  if ( dhCurStrategy::strategy() ) 
    dhCurStrategy::strategy()->configure(this);

  m_subject = 0;

}


//----------------------------------------------------------------------
//  Destruct an option.

dhOption::~dhOption (void) {

}


//----------------------------------------------------------------------
//  Note that an event has occurred.

void dhOption::note (dhEvent *x) {

  dhComponent::note(x);
  if ( x->bubbling() ) return;

  if (m_subject) m_subject->note(x);

  if ( x->tell_parent() && parent() ) parent()->note(x);

}


//======================================================================
//  dhSubProb   -   Base class for decisions and problems.
//
//----------------------------------------------------------------------
//  Construct a subproblem.

dhSubProb::dhSubProb (void) {

  m_strategy = 0;
  if ( dhCurStrategy::strategy() ) 
    m_strategy = dhCurStrategy::strategy();

  m_max_way = m_num_way = 0;
  m_way = 0;

  m_i_this_way = m_i_last_way = m_i_best_way = -1;

}


//----------------------------------------------------------------------
//  Destruct a subproblem.

dhSubProb::~dhSubProb (void) {

  for (short iway=0; iway<m_num_way; iway++) {
    if (m_way[iway]) delete m_way[iway];
  }

  if (m_way) delete m_way;

}


//----------------------------------------------------------------------
//  Note that an event has occurred.

void dhSubProb::note (dhEvent *x) {

  dhComponent::note(x);
  if ( x->bubbling() ) return;

  if ( x->type() == dhEv_This_Best ) {
    m_i_best_way = m_i_this_way;
  }

}


//----------------------------------------------------------------------
//  Return a pointer to the strategy if one is available.

inline dhStrategy * dhSubProb::strategy (void) {

  if (m_strategy) return m_strategy;

  if ( dhCurStrategy::strategy() ) {
    return dhCurStrategy::strategy();
  }

}


//----------------------------------------------------------------------
//  Allocate way array.
//
//  Allocate both the array of pointers and individual ways to
//  which the pointers point.


void dhSubProb::allocate_ways (short num) {

  if (m_max_way > 0) 
    throw aipXcpt ("WayAlready", "Attempt to allocate ways twice.");

  m_way = new dhWay* [num];

  if (!m_way) 
    throw aipXcpt ("SubProbAllcWays", "Error allocating way array.");

  m_max_way = num;

}


//----------------------------------------------------------------------
//  Return a pointer to the 'iway'th way.

dhWay  * dhSubProb::way (short iway) const {

#ifndef dhNoChk
  if (iway < 0 || iway >= m_num_way) {
    throw aipXcpt ("DcsnIway", "Invalid way index");
  }
#endif

  return m_way[iway];

}


//----------------------------------------------------------------------
//  Return a pointer to this-way.

dhWay  * dhSubProb::this_way  (void) const  { 

  if (m_i_this_way < 0) return 0;

  return way(m_i_this_way); 

}


//----------------------------------------------------------------------
//  Return a pointer to last-way (or this-way if no last-way).

dhWay  * dhSubProb::last_way  (void) const  { 

  if (m_i_last_way >= 0) return way(m_i_last_way); 

  return this_way(); 

}


//----------------------------------------------------------------------
//  Return a pointer to best-way.

dhWay  * dhSubProb::best_way  (void) const  { 

  if (m_i_best_way >= 0) return way(m_i_best_way); 

  return this_way(); 

}


//----------------------------------------------------------------------
//  Set m_i_this_way to refer to a way in which the next solution
//  may be recorded.  The best-way will not be overwritten, if one
//  exists.

dhWay *  dhSubProb::next_way (void) {

  if (m_i_this_way < 0) {  // Havn't used any ways yet.

    m_i_this_way = m_i_last_way = 0;

  } else {                 // Have at least one existing way.

    m_i_last_way = m_i_this_way;
    if (++m_i_this_way >= m_num_way) m_i_this_way = 0;
    if (m_i_this_way == m_i_best_way) {
      if (++m_i_this_way >= m_num_way) m_i_this_way = 0;
    }

  }

  return this_way();

}


//----------------------------------------------------------------------
//  Set the maximum number of ways that this subproblem can maintain.

void dhSubProb::set_max_way (short x) {

  allocate_ways(x);

}


//----------------------------------------------------------------------
//  Add a way to this subproblem.

void dhSubProb::add_way (dhWay *x) {

  if (m_max_way <=0) allocate_ways(dhDefMaxWay);

  if (m_num_way == m_max_way) 
    throw aipXcpt ("TooManyWay", "Attempt to add too many ways.");

  m_way[m_num_way++] = x;

}


//======================================================================
//  dhDecision   -   Set of options and ways in a problem.
//
//----------------------------------------------------------------------
//  Construct a decision.

dhDecision::dhDecision (void) {

  if ( dhCurStrategy::strategy() ) 
    dhCurStrategy::strategy()->configure(this);

  m_max_option = m_num_option = 0;
  m_option = 0;

}


//----------------------------------------------------------------------
//  Destruct a decision.

dhDecision::~dhDecision (void) {

  for (short iopt=0; iopt<m_num_option; iopt++) {
    if (m_option[iopt]) delete m_option[iopt];
  }

  if (m_option) delete m_option;

}


//----------------------------------------------------------------------
//  Allocate option array.

void dhDecision::allocate_options (short num) {

  if (m_max_option > 0) 
    throw aipXcpt ("OptAlready", "Attempt to allocate options twice.");

  m_option = new dhOption* [num];

  if (!m_option) throw aipXcpt ("DcsnAllcOpt",
				"Error allocating options.");

  m_max_option = num;

}


//----------------------------------------------------------------------
//  Note that an event has occurred.
//
//  When options are being told, tell_parent must be turned off
//  temporarily because this is the parent.

void dhDecision::note (dhEvent *x) {

  dhSubProb::note(x);
  if ( x->bubbling() ) return;

  bool save_tell_parent = x->set_tell_parent(false);

  if ( x->tell_children() ) {

    for (int i=0; i<m_num_option; i++) {  // Tell all options.
      option(i)->note(x);
    }

  } else {   // Only tell 'this-way' option.

    dhOption *opt = this_way()->option();
    if (opt) opt->note(x);

  }

  x->set_tell_parent (save_tell_parent);

  if ( x->tell_parent() && parent() ) parent()->note(x);

}


//----------------------------------------------------------------------
//  Return a pointer to the 'iopt'th option.

dhOption  * dhDecision::option (short iopt) const {

#ifndef dhNoChk
  if (iopt < 0 || iopt >= m_num_option) {
    throw aipXcpt ("DcsnIopt", "Invalid decision index");
  }
#endif

  return m_option[iopt];

}


//----------------------------------------------------------------------
//  Set the maximum number of options that this decision can have.

void dhDecision::set_max_option (short x) {

  allocate_options(x);

}


//----------------------------------------------------------------------
//  Add an option to this decision.

void dhDecision::add_option  (dhOption  *x) {

  if (m_max_option <=0) allocate_options(dhDefMaxOption);

  if (m_num_option == m_max_option) 
    throw aipXcpt ("TooManyOpt", "Attempt to add too many options.");

  m_option[m_num_option++] = x;

  x->set_decision(this);

}


//----------------------------------------------------------------------
//  If a strategy is available, attempt to make this decision.
//
//  If a way is established, its recorded goodness may reflect 
//  strategic factors.  For this reason, the problem-domain goodness
//  of the way is determined.


bool  dhDecision::solve (void) {

  if ( ! strategy() ) return false;

  next_way();

  if ( ! strategy()->solve(this) ) return false;

  dhWay *way = this_way();
  if (way) {

    way->set_goodness ( evaluate(way) );

    if ( way->option() ) {

      way->add_goodness ( way->option()->evaluate(way) );

      dhSubject *subject = way->option()->subject();
      if (subject) {
	way->add_goodness ( subject->evaluate(way) );
      }

    }  // End of block applying goodness of option.

    dhWayEvent event(dhEv_Dcsn_Made, way);
    event.set_tell_parent(true);
    note(&event);
    
  }  // End of block handling a way.

  return true;
  
}


//======================================================================
//  dhProblem   -   Composite of decisions and/or problems.
//
//----------------------------------------------------------------------
//  Construct a problem.

dhProblem::dhProblem (void) {

  if ( dhCurStrategy::strategy() ) 
    dhCurStrategy::strategy()->configure(this);

  m_max_subprob = m_num_subprob = 0;
  m_subprob = 0;

  m_max_subject = m_num_subject = 0;
  m_subject = 0;

}


//----------------------------------------------------------------------
//  Destruct a problem.

dhProblem::~dhProblem (void) {

  for (short isp=0; isp<m_num_subprob; isp++) {
    if (m_subprob[isp]) delete m_subprob[isp];
  }

  if (m_subprob) delete m_subprob;

  for (short isbj=0; isbj<m_num_subject; isbj++) {
    if (m_subject[isbj]) delete m_subject[isbj];
  }

  if (m_subject) delete m_subject;

}


//----------------------------------------------------------------------
//  Allocate subproblem array.

void dhProblem::allocate_subprobs (short num) {

  if (m_max_subprob > 0) 
    throw aipXcpt ("SubPAlready", 
		   "Attempt to allocate subprobs twice.");

  m_subprob = new dhSubProb* [num];

  if (!m_subprob) throw aipXcpt ("ProbAllcSubprob",
				 "Error allocating subproblems.");

  m_max_subprob = num;

}


//----------------------------------------------------------------------
//  Allocate subject array.

void dhProblem::allocate_subjects (short num) {

  if (m_max_subject > 0) 
    throw aipXcpt ("SbjAlready", 
		   "Attempt to allocate subjects twice.");

  m_subject = new dhSubject* [num];

  if (!m_subject) throw aipXcpt ("ProbAllcSbj",
				 "Error allocating subjects.");

  m_max_subject = num;

}


//----------------------------------------------------------------------
//  Note that an event has occurred.

void dhProblem::note (dhEvent *x) {

  dhSubProb::note(x);   // Parent in class (not problem) hierarchy
  if ( x->bubbling() ) return;

  if (x->type() == dhEv_Dcsn_Made) {

    dhWayEvent *way_event = (dhWayEvent*) x;
    if ( way_event->way() ) {
      this_way()->add_goodness ( way_event->way()->goodness() );
    }

  } else if (x->type() == dhEv_Dcsn_UnMade) {

    dhWayEvent *way_event = (dhWayEvent*) x;
    if ( way_event->way() ) {
      this_way()->subtract_goodness ( way_event->way()->goodness() );
    }

  }

  if ( x->tell_children() ) {
    for (short isp=0; isp<m_num_subprob; isp++) {
      subprob(isp)->note(x);
    }
  }

  if ( x->tell_parent() && parent() ) parent()->note(x);

}


//----------------------------------------------------------------------
//  Return a pointer to the 'isubprob'th subprob of this problem.

dhSubProb *  dhProblem::subprob (short isubprob) const {

#ifndef dhNoChk
  if (isubprob < 0 || isubprob >= m_num_subprob) {
    throw aipXcpt ("ProbIsubprob", "Invalid subprob index");
  }
#endif

  return m_subprob[isubprob];

}


//----------------------------------------------------------------------
//  Return a pointer to the 'isbj'th subject of this problem.

dhSubject *  dhProblem::subject (short isbj) const {

#ifndef dhNoChk
  if (isbj < 0 || isbj >= m_num_subject) {
    throw aipXcpt ("ProbIsbj", "Invalid subject index");
  }
#endif

  return m_subject[isbj];

}


//----------------------------------------------------------------------
//  Set the maximum number of subprobs this problem may have.

void dhProblem::set_max_subprob (short x) {

  allocate_subprobs(x);

}


//----------------------------------------------------------------------
//  Set the maximum number of subjects this problem may have.

void dhProblem::set_max_subject (short x) {

  allocate_subjects(x);

}


//----------------------------------------------------------------------
//  Add a subprob to this problem.

void dhProblem::add_subprob (dhSubProb *x) {

  if (m_max_subprob <=0) allocate_subprobs(dhDefMaxSubProb);

  if (m_num_subprob == m_max_subprob) 
    throw aipXcpt ("TooManySubProb", 
		   "Attempt to add too many subproblems.");

  m_subprob[m_num_subprob++] = x;

  x->set_problem(this);

}


//----------------------------------------------------------------------
//  Add a subject to this problem.

void dhProblem::add_subject (dhSubject *x) {

  if (m_max_subject <=0) allocate_subjects(dhDefMaxSubject);

  if (m_num_subject == m_max_subject) 
    throw aipXcpt ("TooManySbj", "Attempt to add too many subjects.");

  m_subject[m_num_subject++] = x;

  x->set_owner(this);

}


//----------------------------------------------------------------------
//  If a strategy is available, attempt to solve this problem.

bool  dhProblem::solve (void) {

  if ( strategy() ) {

    next_way();

    if ( strategy()->solve(this) ) return true;

  }

  return false;

}


//======================================================================
//  dhSubject   -   Abstract subject of options of decisions.
//
//----------------------------------------------------------------------
//  Construct a subject.

dhSubject::dhSubject (void) {

  if ( dhCurStrategy::strategy() ) 
    dhCurStrategy::strategy()->configure(this);

}


//----------------------------------------------------------------------
//  Destruct a subject.

dhSubject::~dhSubject (void) {

}


//----------------------------------------------------------------------
//  Note that an event has occurred.

void dhSubject::note (dhEvent *x) {

  dhComponent::note(x);
  if ( x->bubbling() ) return;

  if ( x->tell_parent() && parent() ) parent()->note(x);

}


//======================================================================
//  dhWay   -   Choice of an option in a decision.
//
//----------------------------------------------------------------------
//  Construct a way.

dhWay::dhWay (void) {

  m_subprob = 0;
  m_option = 0;

  m_goodness = aipNeutral;

}


//----------------------------------------------------------------------
//  Construct a way given a subproblem, an option and an optional 
//  goodness.

dhWay::dhWay (dhSubProb *sp, dhOption *o, aipGoodness g) {

  m_subprob = sp;
  m_option = o;

  m_goodness = g;

}


//----------------------------------------------------------------------
//  Construct a way as a copy of 'way'.

dhWay::dhWay (dhWay& way) {

  m_subprob  = way.subprob();
  m_option   = way.option();

  m_goodness = way.goodness();

}


//----------------------------------------------------------------------
//  Destruct a way.

dhWay::~dhWay (void) {

}


//----------------------------------------------------------------------
//  Assign 'way' to this way.

dhWay& dhWay::operator = (dhWay& way) {

  m_subprob  = way.subprob();
  m_option   = way.option();

  m_goodness = way.goodness();

}


//======================================================================
//  dhEvent   -   Event relating to a dirty-hard problem.
//
//----------------------------------------------------------------------
//  Construct an event.

dhEvent::dhEvent (dhEventType type) {

    m_type = type; 
    m_tell_children = false;
    m_tell_parent   = false;

}


//----------------------------------------------------------------------
//  Destruct an event

dhEvent::~dhEvent (void) {

}


//----------------------------------------------------------------------
//  Set whether the event is to be told to the children of this
//  object in the problem tree.  Return the value before being set.

bool dhEvent::set_tell_children (bool x) { 

  bool save = m_tell_children;
  m_tell_children = x;
  return save;

}

//----------------------------------------------------------------------
//  Set whether the event is to be told to the parent of this
//  object in the problem tree.  Return the value before being set.

bool dhEvent::set_tell_parent (bool x) { 

  bool save = m_tell_parent;
  m_tell_parent = x;
  return save;

}

//======================================================================
//  dhStrategy   -   Abstract strategy for solving dirty-hard problems.
//
//----------------------------------------------------------------------
//  Configure 'x' with a strategy-aspect, if required by this strategy.

void dhStrategy::configure (dhProblem *x) {

  dhAspect *aspect = create_problem_aspect();
  if (aspect) x->add_strategy_aspect(aspect);

  configure_ways(x);

}


//----------------------------------------------------------------------
//  Configure 'x' with a strategy-aspect and/or ways, as required 
//  by this strategy.

void dhStrategy::configure (dhDecision  *x) {

  dhAspect *aspect = create_decision_aspect();
  if (aspect) x->add_strategy_aspect(aspect);

  configure_ways(x);

}


//----------------------------------------------------------------------
//  Configure 'x' with a strategy-aspect, if required by this strategy.

void dhStrategy::configure (dhOption *x) {

  dhAspect *aspect = create_option_aspect();
  if (aspect) x->add_strategy_aspect(aspect);

}


//----------------------------------------------------------------------
//  Configure 'x' with a strategy-aspect, if required by this strategy.

void dhStrategy::configure (dhSubject *x) {

  dhAspect *aspect = create_subject_aspect();
  if (aspect) x->add_strategy_aspect(aspect);

}


//----------------------------------------------------------------------
//  Configure this subproblem with ways.

void dhStrategy::configure_ways (dhSubProb *x) {

  short nway = num_way();

  if (nway <= 0) return;

  x->set_max_way(nway);

  for (int iway=0; iway<nway; iway++) {
    dhWay *way = create_way();
    if (!way) throw aipXcpt ("WayCreate", "Error creating way.");
    way->set_subprob(x);
    x->add_way(way);
  }
  
}


//======================================================================
//  dhCurStrategy   -   The 'current' strategy.
//
//----------------------------------------------------------------------
//  Definition of static data member.

dhStrategy * dhCurStrategy::m_strategy = 0;


//----------------------------------------------------------------------
//  Static function to set the current strategy.

void  dhCurStrategy::set_strategy (dhStrategy  *x) { 

  m_strategy = x; 

}


//----------------------------------------------------------------------
//  Static function to return the current strategy.

dhStrategy  * dhCurStrategy::strategy (void) { 

  return m_strategy; 

}


//======================================================================
//                           License
//
//   Permission is hereby granted, free of charge, to any 
//   person obtaining a copy of this software and associated 
//   documentation files (the "Software"), to deal in the Software 
//   without restriction, including without limitation the rights 
//   to use, copy, modify, merge, publish, distribute, sublicense, 
//   and/or sell copies of the Software, and to permit persons to 
//   whom the Software is furnished to do so, subject to the 
//   following conditions:
//
//   The copyright notice and this license shall be included in all 
//   copies or substantial portions of the Software.
//
//   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
//   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES 
//   OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
//   NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
//   HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
//   WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
//   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
//   OTHER DEALINGS IN THE SOFTWARE.
//
//======================================================================

