//**********************************************************************
//  dirthard.h   -   Dirty Hard Problems
//
//  This header file declares...
//    - dirty-hard problems, decisions, options, subjects
//    - aspects, ways, events, strategies, factories
//
//  This software throws aipXcpt exceptions on problems.
//
//  Copyright Brian Marshall 1999
//
//  See the license at end of this file.
//
//  Brian Marshall    bmarshal@agt.net    (403)651-0584
//
//  99/03/06  Brian Marshall  development begun.
//  99/05/01  Brian Marshall  problems as composites, major reorg.
//  99/06/06  Brian Marshall  current and subproblem strategies.
//  99/06/23  Brian Marshall  work on events, ways, comments.
//
//
//----------------------------------------------------------------------

#ifndef dirthard_h_guard
#define dirthard_h_guard

                           // Goodness, importance, likelihood.
#include "aipGood.h"
                           // Exceptions.
#include "aipXcpt.h"


//----------------------------------------------------------------------
//  Classes and types.  Sub-classing is shown by indentation.

class  dhBase;               // Parts that respond to events
  class  dhAspect;           // Aspect in a problem.
  class  dhComponent;        // Base class for problem components.
    class  dhOption;         // Possibility in decision; with aspect.
    class  dhSubProb;        // Base class for decisions and problems.
      class  dhDecision;     // List of options from which to choose.
      class  dhProblem;      // Composite of decisions and problems.
    class  dhSubject;        // Subject of an option of a decision.

class  dhWay;                // Choice of an option in a decision.

class  dhEvent;              // Event relating to a dirty-hard problem.
  class  dhWayEvent;         // Event that specifies a way.

class  dhFactory;            // Factory for building a problem.

class  dhStrategy;           // Abstract problem-solving strategy.

class  dhCurStrategy;        // Current strategy visible everywhere.

//----------------------------------------------------------------------
//  Event Type Codes

typedef short int dhEventType;
                                             // Event created by:
const dhEventType  dhEv_Dcsn_Made    =   1;  // dhDecsion.solve()
const dhEventType  dhEv_Dcsn_UnMade  =   2;  // a strategy's solve()
const dhEventType  dhEv_This_Best    =   3;  // a strategy's solve()


//======================================================================
//  Dirty-Hard Problems
//
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//  Problems, Ways and Solutions
//
//  A dirty-hard problem is represented as a tree:
//    - A problem is a composite of sub-problems and/or decisions
//      - A decision has options
//        - An option may refer to a subject
//
//  More than one option may refer to the same subject.
//  (Tree branches may join at the tip.)
//
//  Problems, decisions, options and subjects are 'components'.
//  A component has pointers to two aspects: one specific to the
//  problem domain and one specific to a strategy.  These pointers
//  may or may not be used.
//
//  Components and aspects descend from dhBase, and have a virtual
//  function  note()  to which an event may be passed.
//
//  Problems and decisions each have an array of 'way's.  For a 
//  decision, a way specifies a choice of an option and the goodness 
//  of that choice.  For a problem, a way specifies the sum of the
//  goodness of the corresponding way in each of its sub-problems.
//
//  Components have indecies to indicate which of its ways is:
//    this-way  -  the way being tried or last tried
//    last-way  -  the way tried before this-way
//    best-way  -  the best way tried so far
//
//  The ways and the indecies of a component are set by its solve()
//  function which calls the solve() function of a strategy object.
//
//  A solution is composed of the options identified in the best-way
//  of each decision.
//
//
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//  Specialization for a Problem Domain
//
//  These classes must be specialized to represent problems in a
//  particular problem domain like staff-scheduling or chess.  This
//  generally involves:
//
//   - subclasses of components that:
//      - provide new names for objects and functions 
//      - assist in building the problem and retrieving the solution
//
//   - subclasses of dhAspect to store domain-specific information
//
//   - a subclass of dhFactory to build problems
//
//  Domain-specific aspects may be attached to any type of component.
//  If the only reason to subclass a component is to add data-members,
//  an aspect can be subclassed instead.
//
//
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//  Strategy Objects
//
//  A strategy is an object that can solve problems and make decisions.
//  A strategy may be implemented in a subclass of dhStrategy.
//
//  Strategies can be independent of the problem-domain.  Strategies 
//  can work with the problem at the level of the classes defined in 
//  this file.  On the other hand, strategies for particular problem
//  domains can also be developed.
//
//  A strategy may want to store data in components.  Subclasses of 
//  dhAspect can be declared.  Strategy-specific aspects may be 
//  attached to any type of component.
//
//  The class dhCurStrategy has:
//    - a static pointer to a 'Current' strategy 
//    - static functions to set and get the pointer  
//  There is no requirement to use this facility.  It is an
//  optional non-exclusive singleton.
//
//  If a 'Current' strategy exists, the constructor of any component
//  will ask the strategy to configure the component with an aspect.
//  A factory can create a strategy, make it the 'Current' strategy,
//  and then build a problem without regard to what the strategy
//  requires.
//
//  Problems and decisions each have a pointer to a strategy.  A
//  strategy is considered to be 'available' to a problem or decision
//  if this pointer points to a strategy or there is a 'Current'
//  strategy.
//
//  If a 'Current' strategy exists, the constructor of any component
//  will set the component's strategy pointer to this strategy.
//
//  A strategy available to a decision may ask the decision to 
//  evaluate a way, a choice of an option.  If the decision has a 
//  problem aspect, this aspect is asked to evaluate the way, 
//  otherwise the evaluation is considered to be aipNeutral.  The
//  results of the evaluation, a goodness, is stored in the way.
//
//
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//  Events 
//
//  An event, represented by a dhEvent object, is notification that
//  something has happened -
//    - a decision has been made or unmade
//    - a solution has been recognized as being the best so far
//    - other events defined for problem-domains and strategies
//
//  if  m_tell_parent  is true:
//    - components (other than the root of the problem tree) tell
//      their parent (the event propogates up the tree)
//
//  If  m_tell_children  is true:
//    - problems tell their sub-problems and decisions
//    - decisions tell all their options (not just the chosen one)
//    (the event propogates to all children below the component)
//
//  If both  m_tell_parent  and  m_tell_children  are true:
//    - the event is 'bubbled' up to the root of the problem tree
//    - m_tell_parent is set to false
//    - the event is told to all children, recursively.
//    (the event propogates to every component in the problem)
//
//  When a component notes an event (other than during 'bubbling'),
//  it tells its problem-domain specific aspect and its strategy
//  specific aspect.
//
//  When an option notes an event, it will tell its subject, if it
//  has one.
//
//  
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//  Problem-Solving Applications
//
//  Applications will generally:
//   - construct a factory and a strategy
//   - make the strategy the current strategy
//   - use the factory to construct problems, decisions, options
//     and subjects (which will be configured by the strategy).
//   - hook these components together to create a problem tree
//   - call solve() on the root problem of the tree
//   - iterate through the tree, retrieving the best-way
//
//  Well-tested implementations used in applications may be compiled
//  with  dhNoChk  defined to improve performance.  This will supress
//  bounds checking in functions returning an pointer to an object
//  that are passed an index.
//
//
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//  Design Notes
//
//  This code should be able to be compiled with any C++ compiler
//  that supports exceptions.
//
//  Simple pointer-arrays are used instead of standard library 
//  containers because some people are using compilers that don't 
//  support containers well.  Similarily, 'C'-style casting is used
//  for down-casting.
//
//  We use constant short ints for event type codes rather than an 
//  enumeration so that if an event-type is written to a file and a 
//  new event-type is defined, the event-type number in the file will 
//  continue to refer to the same event.
//
//
//======================================================================
//  dhBase   -   Base class for components and aspects.


class dhBase {

public:

  dhBase (void) {}
  virtual ~dhBase (void) {}

  virtual void note (dhEvent *x) {}

};


//======================================================================
//  dhAspect   -   Aspect in a problem.
//
//  Subclasses of dhAspect may be attached to components to store 
//  problem-domain or strategic information.   Subclasses may
//  override 'evaluate()' to determine the current goodness associated 
//  with the aspect, store it in m_g and return it.


class dhAspect : public dhBase {

  aipGoodness   m_g;     // Current (or latest) results of evaluate.

public:

  dhAspect (void) { m_g = aipNeutral; }
  virtual ~dhAspect (void) {}

  aipGoodness goodness (void) const { return m_g; }

  void set_goodness (aipGoodness x) { m_g = x; }

  virtual aipGoodness evaluate (dhWay *way) { return m_g; }

};


//======================================================================
//  dhComponent   -   Component of a dirty-hard problem tree.
//
//  Problems, decisions, options and subjects are dhComponents.
//
//  A problem-aspect might be added to a component by a subclass of
//  dhFactory specialized for a particular problem domain.  
//
//  The constructors of classes descending directly from dhComponent 
//  must give the strategy a chance to configure the component with
//  an aspect particular to the strategy, if a 'current' strategy
//  exists.
//
//  Aspects are deleted when a component is destroyed.  If a 
//  component already has an aspect when add_xxx_aspect is called,
//  the replaced aspect is not deleted.
//
//  Subclasses will provide a function that uses parent() and
//  returns a pointer to a more specific type.  Similarily, they
//  will provide a function to add a parent that uses  set_parent()
//  that takes an argument of a more specific type.
//
//  In subclasses that override note(), note() will:
//   - call dhComponent::note()
//   - if event.bubbling(): return
//   - call note() for any member aspects
//   - if event.tell_children(): call note() for children


class dhComponent : public dhBase {

  dhComponent  * m_parent;            // Parent node in tree.

  dhAspect     * m_problem_aspect;    // Aspect for problem-domain.
  dhAspect     * m_strategy_aspect;   // Aspect for strategy.

public:

  dhComponent (void);
  virtual ~dhComponent (void);

  virtual void note (dhEvent *x);

  dhComponent *  parent (void) const { return m_parent; }

  bool is_root (void) const { return ! m_parent; }

  dhAspect * problem_aspect  (void) const { return m_problem_aspect;  }
  dhAspect * strategy_aspect (void) const { return m_strategy_aspect; }

  void set_parent (dhComponent *x) { m_parent = x; }

  void add_problem_aspect  (dhAspect *x) { m_problem_aspect  = x; }
  void add_strategy_aspect (dhAspect *x) { m_strategy_aspect = x; }

  aipGoodness evaluate (dhWay *way);

};


//======================================================================
//  dhOption   -   Possibility in a decision.
//
//  Options may be configured to refer to a (subclass of a)
//  subject.  The subject is not owned by the option.


class dhOption : public dhComponent {

  dhSubject     * m_subject;           // Subject of option or null.

public:

  dhOption (void);
  virtual ~dhOption (void);

  virtual void note (dhEvent *x);

  dhDecision * decision (void) const  { return (dhDecision*)parent(); }
  dhSubject  * subject  (void) const  { return m_subject;  }

  void set_decision  (dhDecision *x) { set_parent((dhComponent*)x); }

  void set_subject   (dhSubject  *x) { m_subject = x; }

};


//======================================================================
//  dhSubProb   -   Base class for decisions and problems.
//
//  A subproblem is something that must be solved for an overall 
//  problem to be solved.  How this is done in a particular case is
//  recorded in a way.
//
//  If a current strategy has been set at the time a subproblem is
//  constructed, the current strategy becomes the strategy for the
//  the subproblem.  A strategy is considered to be 'available' for
//  a subproblem if it has a strategy or, failing that, there is a
//  current strategy.  Subproblems do not own strategies.
//
//  The default maximum number of ways is defined below the 
//  class declaration.  This can be changed once with 
//  set_max_way() before any ways are added to the array.
//
//  next_way() swaps the pointers to this-way and last-way, and
//  returns a pointer to the new this-way.


class dhSubProb : public dhComponent {

  dhStrategy * m_strategy;        // Strategy for solving.

  short        m_max_way;         // Num array elements.
  short        m_num_way;         // Num pointers in array.
  dhWay     ** m_way;             // Array of pointers to ways.

  short        m_i_this_way;      // Index of current way.
  short        m_i_last_way;      // Index of current way.
  short        m_i_best_way;      // Index of best way.

  void allocate_ways    (short num);   // Allocate array of ways.

 public:

  dhSubProb (void);
  virtual ~dhSubProb (void);

  virtual void note (dhEvent *x);

  dhStrategy * strategy (void);

  dhProblem * problem (void) const  { return (dhProblem*)parent();  }
  short       max_way      (void) const  { return m_max_way;        }
  short       num_way      (void) const  { return m_num_way;        }

  dhWay     * way       (short i) const;

  inline dhWay  * this_way  (void) const;
  inline dhWay  * last_way  (void) const;
  inline dhWay  * best_way  (void) const;

  dhWay  * next_way (void);

  void set_problem  (dhProblem  *x) { set_parent((dhComponent*)x); }
  void set_strategy (dhStrategy *x) { m_strategy = x;              }
  void set_max_way  (short x);

  void add_way (dhWay *x);

  virtual bool solve (void) =0;

};


// Default maximum ways per object...
const short  dhDefMaxWay  =  3;


//======================================================================
//  dhDecision   -   Set of options and ways in a problem.
//
//  A decision has a list of options.  Solving a decision involves
//  choosing one of the options.  A way records such a choice.
//  The following ways are defined:
//     'this-way'   the way being set or that was last set. 
//     'last-way'   the way done immediately before this-way. 
//     'best-way'   this-way when dhEv_This_Best event last noted.
//
//  A strategy may define additional ways (other, different, that,
//  worst, etc.).  The number of ways in a decision is generally 
//  controlled by a strategy.
//
//  If solve() is called and a strategy is available:
//   - it will be asked to choose an option
//   - a pointer to the option and the goodness returned by the 
//     option's problem-aspect are stored in a way
//   - this decision is asked to note the way
//
//  When a decision notes a dhEvent_Dcsn_Made, it sends the event
//  to the chosen option and to its parent problem.  (The option
//  will send the event to the subject, if there is one.)
//
//  The default maximum number of options is defined below the 
//  class declaration.  This can be changed once with 
//  set_max_option() before any options are added to the array.


class dhDecision : public dhSubProb {

  short        m_max_option;      // Num array elements.
  short        m_num_option;      // Num pointers in array.
  dhOption  ** m_option;          // Array of pointers to options.

  void allocate_options (short num);   // Allocate array of options.

public:

  dhDecision (void);
  virtual ~dhDecision (void);

  virtual void note (dhEvent *x);

  short       max_option   (void) const  { return m_max_option;     }
  short       num_option   (void) const  { return m_num_option;     }

  dhOption  * option    (short i) const;

  void set_max_option (short x);

  void add_option  (dhOption  *x);

  virtual bool solve (void);

};


// Default maximum options per decision...
const short  dhDefMaxOption   =  8;


//======================================================================
//  dhProblem   -   Composite of decisions and/or problems.
//
//  A problem is a set of subproblems that must be solved.  It is
//  a composite of decisions and problems.  The default maximum 
//  number of subproblems is defined below the class declaration.  
//  This can be changed once with set_max_subproblem() before the 
//  first subproblem is added.
//
//  If solve() is called and a strategy is available, it will be 
//  asked to solve this problem - 'solve' all its decisions.
//
//  A problem is a dhSubProb, and has an array of ways and 
//  indecies indicating which is this-way, last-way and best-way.
//  A way in a problem contains the sum of the goodness of the
//  corresponding way in the subproblems.  For instance, the way 
//  pointed to by best_way() contains the sum of 
//  best_way()->goodness() for each subproblem in the problem.
//
//  When a problem notes a dhEvent_Dcsn_Made, it adds the goodness
//  in the way specified in the event to the goodness in this
//  problem's this-way.  It will then pass the event to its parent
//  problem, if there is one.
//
//  A problem may own a list of subjects - objects that are the
//  subject of options in decisions.
//
//  If m_array_order is true, the subproblems in the problem must
//  be solved in their order in m_decision.  If m_array_order
//  is false (the default), the strategy is free to address
//  subproblems in any order it chooses.


class dhProblem : public dhSubProb {

  short          m_max_subprob;   // Num array elements.
  short          m_num_subprob;   // Num pointers in array.
  dhSubProb   ** m_subprob;       // Array of pointers to 
                                     // decisions and problems.

  short          m_max_subject;    // Num array elements.
  short          m_num_subject;    // Num pointers in array.
  dhSubject   ** m_subject;        // Array of pointers to subjects.

  bool           m_array_order;    // Solve in array order.

  void allocate_subprobs  (short num);   // Allocate array.
  void allocate_subjects  (short num);   // Allocate array.

public:

  dhProblem (void);
  virtual ~dhProblem (void);

  virtual void note (dhEvent *x);

  short  max_subprob  (void)  const { return m_max_subprob; }
  short  max_subject  (void)  const { return m_max_subject; }
  short  num_subprob  (void)  const { return m_num_subprob; }
  short  num_subject  (void)  const { return m_num_subject; }
  bool   array_order  (void)  const { return m_array_order; }

  dhSubProb  * subprob  (short i) const;
  dhSubject  * subject  (short i) const;

  void set_max_subprob (short x);
  void set_max_subject (short x);

  void add_subprob (dhSubProb *x);
  void add_subject (dhSubject *x);

  void set_array_order (bool x) { m_array_order = x; }

  virtual bool solve (void);

};


//  Default max subproblems and subjects per problem...
const short  dhDefMaxSubProb  =  128;
const short  dhDefMaxSubject  =  128;


//======================================================================
//  dhSubject   -   Abstract subject of options of decisions.
//
//  In a problem, there may be objects that may be the subject of
//  more than one decision or option, and it may be desirable to
//  store information about them.  For example, in a staff schedule,
//  decisions are made for employees in relation to shifts.  The
//  dhSubject class could be subclassed for shifts.  
//
//  A subject is a dhComponent - it can note events and has aspects.
//  Subjects are referred to by options.  A subject may be the subject
//  of more than one option from more than one decision.  The parent
//  of a subject is the problem that owns it.  (This is not 
//  necessarily the problem with the decisions with the referring
//  options; it may be a parent of that problem.)


class dhSubject : public dhComponent {

public:

  dhSubject (void);
  virtual ~dhSubject (void);

  virtual void note (dhEvent *x);

  dhSubProb * owner (void) const  { return (dhSubProb*)parent(); }

  void set_owner  (dhProblem  *x) { set_parent((dhComponent*)x); }

};


//======================================================================
//  dhWay   -   What was (or could be) done.
//
//  For a way attached to a decision:
//    - m_option may point to a chosen option
//    - m_goodness is the goodness of that choice
//
//  For a way attached to a problem:
//    - the goodness is the sum of the goodnesses in the corresponding 
//      ways of its decisions.
//    - m_option is not used.
//
//  A way may be passed to its decision and ask to be evaluated.
//  This will be done by the decision's problem-aspect if one exists.
//
//  A way knows about a decision and an option; it does not own them.


class dhWay {

  dhSubProb    * m_subprob;     // SubProb affected.
  dhOption     * m_option;      // The chosen option or null.
  aipGoodness    m_goodness;    // Goodness of this way.

public:

  dhWay (void);
  dhWay (dhSubProb *sp, dhOption *o, aipGoodness g =aipNeutral);
  dhWay (dhWay& way);

  virtual ~dhWay (void);

  dhWay& operator = (dhWay& x);

  dhSubProb    * subprob   (void) const { return m_subprob;  }
  dhOption     * option    (void) const { return m_option;   }
  aipGoodness    goodness  (void) const { return m_goodness; }

  dhSubject    * subject   (void) const { 
    return m_option ? m_option->subject() : 0;
  }

  void set_subprob   (dhSubProb  *x)   { m_subprob  = x; }
  void set_option    (dhOption   *x)   { m_option   = x; }
  void set_goodness  (aipGoodness x)   { m_goodness = x; }

  void add_goodness      (aipGoodness x)  { m_goodness += x; }
  void subtract_goodness (aipGoodness x)  { m_goodness -= x; }

};


//======================================================================
//  dhEvent   -   Event relating to a dirty-hard problem.
//
//  When an object receives an event, it will tell any aspects it has.
//  By default, children will be told and the parent will not.
//
//  If both parents and children are to told, the event will 
//  bubble up through parents until a component is reached with
//  no parent and then the event will be adjusted so that parents
//  will not be told.  If a component is given an event to note
//  and the event is being bubbled, it should do nothing with the
//  event but call dhComponent::note().


class dhEvent {

  dhEventType  m_type;

  bool         m_tell_children;
  bool         m_tell_parent;

public:

  dhEvent (dhEventType type);
  virtual ~dhEvent (void);

  dhEventType  type (void) const { return m_type; }

  bool tell_children (void) const { return m_tell_children; }
  bool tell_parent   (void) const { return m_tell_parent;   }

  bool set_tell_children (bool x);
  bool set_tell_parent   (bool x);

  bool bubbling (void) { 
    return m_tell_children && m_tell_parent; 
  }

};

//----------------------------------------------------------------------
//  dhWayEvent   -   Events that specify a way.


class dhWayEvent : public dhEvent {

  dhWay   * m_way;

public:

  dhWayEvent (dhEventType type, dhWay *way) : dhEvent(type) {
    m_way = way; 
  }

  virtual ~dhWayEvent (void) {}

  dhWay * way (void) const { return m_way; }

};


//======================================================================
//  dhStrategy   -   Abstract strategy for solving dirty-hard problems.
//
//  Applications may use a subclass of dhStrategy to solve problems
//  constructed with the classes in this file.
//
//  Functions are provided to configure problem-components with
//  strategy-specific aspects.  These functions may be called by
//  the constructors of the components.  These functions call the
//  appropriate create_xxx_aspect() function, and if an aspect is
//  returned, attach it to the passed component.  
//
//  The configure function for decisions also:
//   - calls num_ways(), and if it returns a non-negative number, 
//     resets the number of ways stored by the decision.
//   - creates the required ways and attaches them to the decision.
//
//  Subclasses of dhStrategy will override the create_xxx_aspect()
//  functions for any required aspects.  Note that a problem is a
//  subclass of decision, so create_problem_aspect() returns a
//  pointer to a dhDcsnAspect.
//
//  By default, a strategy expects decisions and problems to be 
//  configured with 3 ways: this-way, last-way and best-way.
//  A subclass of dhStrategy might override num_way() and maintain
//  ways in decision-aspects to implement such concepts as that-way, 
//  other-way, different-way, second-way, worst-way, etc.
//
//  If  solve(dhDecision*)  is overridden, it is expected to:
//   - pick an option that is attached to the decision
//   - record the choice in the decision's this-way
//
//  If  solve(dhProblem*)  is overridden, it is expected to:
//   - solve all its subproblems (decisions and/or problems)
//   - when a best-so-far solution is found, let the problem
//     note a dhEv_This_Best event.


class dhStrategy {

protected:

  virtual dhAspect  * create_problem_aspect  (void) { return 0; }
  virtual dhAspect  * create_decision_aspect (void) { return 0; }
  virtual dhAspect  * create_option_aspect   (void) { return 0; }
  virtual dhAspect  * create_subject_aspect  (void) { return 0; }

  virtual short num_way (void) { return dhDefMaxWay; }
  virtual dhWay  * create_way  (void) { return new dhWay; }

public:

  dhStrategy (void) {}
  virtual ~dhStrategy (void) {}

  void configure (dhProblem   *x);
  void configure (dhDecision  *x);
  void configure (dhOption    *x);
  void configure (dhSubject   *x);

  void configure_ways (dhSubProb *x);

  virtual bool  solve  (dhDecision *x) { return false; }
  virtual bool  solve  (dhProblem  *x) { return false; }

};


//======================================================================
//  dhCurStrategy   -   The 'current' strategy.
//
//  This class provides static functions that make a 'current'
//  strategy visible to everyone.  (There is no need to actually
//  create an object of this class.)
//
//  If a current strategy is set, the constructors of direct 
//  descendents of dhComponent will give the strategy a chance to
//  configure the component with a strategy-specific aspect.


class dhCurStrategy {

  static dhStrategy  * m_strategy;

 public:

  static void set_strategy (dhStrategy *x);

  static dhStrategy  * strategy (void);

};


//======================================================================
//  dhFactory   -   Abstract factory for building dirty-hard problems.
//
//  Applications will generally create a factory that is a subclass
//  of dhFactory.  Such subclasses will provide functions to create
//  objects in a particular problem-domain.
//
//  set_strategy() sets the current strategy (its just an alternate
//  interface to a static function call).


class dhFactory {

public:

  dhFactory (void) {}
  virtual ~dhFactory (void) {}

  void set_strategy (dhStrategy *x) {
    dhCurStrategy::set_strategy(x); 
  }

};


//======================================================================

#endif


//----------------------------------------------------------------------
//                           License
//
//   Permission is hereby granted, free of charge, to any 
//   person obtaining a copy of this software and associated 
//   documentation files (the "Software"), to deal in the Software 
//   without restriction, including without limitation the rights 
//   to use, copy, modify, merge, publish, distribute, sublicense, 
//   and/or sell copies of the Software, and to permit persons to 
//   whom the Software is furnished to do so, subject to the 
//   following conditions:
//
//   The copyright notice and this license shall be included in all 
//   copies or substantial portions of the Software.
//
//   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
//   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES 
//   OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
//   NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
//   HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
//   WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
//   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
//   OTHER DEALINGS IN THE SOFTWARE.
//
//**********************************************************************

