//**********************************************************************
//  roomapp.cxx   -   Application to assign students to rooms.
//
//  This program is a sample application that represents and attempts
//  to solve dirty-hard problems.  
//
//  This program uses  roomprob.h  which specializes  dirthard.h 
//  to solve 'room' problems - assigning students to 'rooms'.  
//
//  Everything to do with input/output files and identifying students
//  and rooms with ids is implemented in this file, not in roomprob.h.
//
//  This program uses  picfirst.h  which is a sample subclass of
//  dhStrategy (declared in dirthard.h).  This is a sample strategy.
//  It is a pretty bad strategy.  It simply picks the first option
//  in each decision.  It is an example of how to code and use a
//  strategy.   >>  Do not expect it to be successful.  << 
//
//  Copyright Brian Marshall 1999
//
//  See the license at end of this file.
//
//  99/05/20  Brian Marshall  development begun
//
//
//----------------------------------------------------------------------
//  The Problem Domain and Interface.
//
//  Given:
//
//    -  some rooms, each with an upper limit on the number of 
//       students it can hold, and,
//
//    -  some students that each have to be assigned to one or 
//       more rooms and that each have preferences,
//
//  assign each student to the required number of rooms.
//
//  The primary purpose of this application is to assign students to
//  optional-classes.  Unfortunately, the words 'option' and 'class'
//  both have special meanings, so we went with the word 'room'.
//
//  The number of rooms to which each student is assigned, and the
//  number of preferences, can vary from student to student.  The 
//  number of students that each room can hold can vary from room 
//  to room.
//
//  The user specifies a filename prefix.  If this is "abc"...
//    - Rooms     are defined according to the file    abc.room
//    - Students  are defined according to the file    abc.student
//    - The solution is written to                     abc.solution
//      (as well as standard output)
//
//  See the sample files beginning with "grade9"
//
//----------------------------------------------------------------------

#include "roomprob.h"
#include "picfirst.h"

#include <fstream.h>

//======================================================================
//  Classes that are specialized for this application.


class appRoom : public rmRoom {
  short   m_id;
public:
  short  id (void) const { return m_id; }
  void  set_id (short x) { m_id = x; }
};


class appStudent : public rmStudent {
  short   m_id;
public:
  appRoom * room (short i) const { return (appRoom*)rmStudent::room(i); }
  short       id (void)        const { return m_id;               }
  short  room_id (short idcsn) const { return room(idcsn)->id();  }
  void  set_id (short x) { m_id = x; }
};


class appProblem : public rmProblem {
public:
  appRoom * room(short i) { 
    return (appRoom*)rmProblem::room(i);
  }
  appStudent * student(short i) { 
    return (appStudent*)rmProblem::student(i);
  }
  void write_solution (char* probfile);
};


class appFactory : public rmFactory {
  void skip_remarks (ifstream& ifs);
  void add_room     (ifstream& ifs);
  void add_student  (ifstream& ifs);
  void add_option_by_id (short room_id);
protected:
  virtual rmProblem  * new_problem  (void) { return new appProblem; }
  virtual rmRoom     * new_room     (void) { return new appRoom;    }
  virtual rmStudent  * new_student  (void) { return new appStudent; }
public:
  appProblem * new_problem (char* file_prefix);
  appProblem * problem (void) { 
    return (appProblem*)rmFactory::problem();
  }
};


//======================================================================
//  Mainline.


int main (int argc, char *argv[]) {

  if (argc != 2) {
    cout << "Usage:  roomapp  file_prefix\n\n"
	 << "  Ex.   roomapp  grade9\n\n"
	 << "     Will read   grade9.room  and \n"
	 << "                 grade9.student\n"
	 << "      and write  grade9.solution\n\n";
    return 1;
  }

  char *file_prefix = argv[1];

  try {

    appFactory  factory;

    PickFirst  strategy;

    factory.set_strategy(&strategy);

    appProblem  *problem = factory.new_problem (file_prefix);

    if ( problem->solve() ) {

      aipGoodness g = problem->best_way()->goodness();

      cout << "The goodness of the best solution found was "
	   << g.val() << " which is " << g.description() << ".\n";

      if (g > aipForbidden) problem->write_solution (file_prefix);

    }  else {

      cout << "A solution was not found.\n";

    }

    delete problem;

    exit(0);

  }

  catch (aipXcpt xcpt) {
    xcpt.describe(cout);
    exit(1);
  }

}


//======================================================================
//  appProblem  -  Member functions
//
//----------------------------------------------------------------------
//  Write this solution to a file, one line per student.

void appProblem::write_solution (char* file_prefix) {

  char outfile[200];
  strcpy (outfile, file_prefix);
  strcat (outfile, ".solution");

  ofstream ofs (outfile);
  if (!ofs) {
    aipXcpt xcpt ("appFileOpen", "Error opening output file: ");
    xcpt.desc_cat (outfile);
    throw xcpt;
  }

  ofs  << "# This is a sample output file. \n\n"
       << "#Student-   Num-   Rooms\n"
       << "#Id         Room   Assigned\n\n";

  cout << "\n\n------------------------------------\n\n"
       << "#Student-   Num-   Rooms\n"
       << "#Id         Room   Assigned\n\n";

  for (short istudent=0; istudent<num_student(); istudent++) {
    appStudent *the_student = student(istudent);

    short num_dcsn = the_student->num_subprob();

    ofs  << "  " << the_student->id()
	 << "         " <<  num_dcsn << "    ";

    cout << "  " << the_student->id()
	 << "         " <<  num_dcsn << "    ";

    for (short idcsn=0; idcsn<num_dcsn; idcsn++) {
      ofs  << "  " << the_student->room_id(idcsn);
      cout << "  " << the_student->room_id(idcsn);
    }

    ofs  << "\n";
    cout << "\n";

  }

  ofs.close();

  cout << "Solution written to: " << outfile << "\n";

}


//======================================================================
//  appFactory  -  Member functions
//
//----------------------------------------------------------------------
//  Skip records which are remarks in ifs.  Remarks are records
//  beginning with '#'.

void appFactory::skip_remarks (ifstream& ifs) {

  char c;

  while ( (ifs >> c) && c == '#' ) {
    while ( ifs.get(c) ) if (c == '\n') break;
  } 

  if (ifs) ifs.putback(c);

}


//----------------------------------------------------------------------
//  Add a new room to the problem according to the next record 
//  in 'ifs'.  If 'ifs' is at eof, return without doing anything.

void appFactory::add_room (ifstream& ifs) {

  skip_remarks (ifs);

  short id, max_student;

  ifs >> id >> max_student;

  if ( ifs.eof() ) return;

  if (!ifs) throw aipXcpt ("appReadRoom", "ERROR reading room.");

  cout << "  id: " << id
       << "     Max students: " << max_student << "\n";

  rmFactory::add_room (max_student);

  ((appRoom*)added_room())->set_id (id);

}


//----------------------------------------------------------------------
//  Add a new student to the problem according to the next record 
//  in 'ifs'.  If 'ifs' is at eof, return without doing anything.

void appFactory::add_student (ifstream& ifs) {

  skip_remarks (ifs);

  short id, num_decision, num_option;

  ifs >> id >> num_decision >> num_option;

  if ( ifs.eof() ) return;

  if (!ifs) throw aipXcpt ("appReadStudent", 
			   "ERROR reading student.");

  cout << "  id: " << id
       << "     Num decisions: " << num_decision
       << "     Num options: "   << num_option << "\n";

  rmFactory::add_student (num_decision, num_option);

  ((appStudent*)added_student())->set_id(id);

  short room_id;

  cout << "    Options:  ";

  for (short iopt=0; iopt<num_option; iopt++) {

    if ( ! (ifs >> room_id) ) {
      throw aipXcpt ("appReadOption", "ERROR reading student option.");
    }

    add_option_by_id (room_id);

    cout << "  " << room_id;

  }

  cout << "\n";

}


//----------------------------------------------------------------------
//  Add a room option, specified as a room-id, to the last student 
//  added to the problem.

void appFactory::add_option_by_id (short room_id) {

  for (int iroom=0; iroom<problem()->num_room(); iroom++) {

    if (problem()->room(iroom)->id() == room_id) {
      add_option(iroom);
      break;
    }

  }  // End of loop looking for room with specified id.

}


//----------------------------------------------------------------------
//  Create a new room-problem from input files.

appProblem  * appFactory::new_problem (char* file_prefix) {

  begin_new_problem();

  char infile[200];

  strcpy (infile, file_prefix);
  strcat (infile, ".room");

  ifstream ifs_room (infile);
  if (!ifs_room) {
    aipXcpt xcpt ("appFileOpen", "Error opening input file: ");
    xcpt.desc_cat (infile);
    throw xcpt;
  }

  cout << "\nRooms:\n";

  while (ifs_room) add_room(ifs_room);

  ifs_room.close();

  strcpy (infile, file_prefix);
  strcat (infile, ".student");

  cout << "\nStudents:\n";

  ifstream ifs_student (infile);
  if (!ifs_student) {
    aipXcpt xcpt ("appFileOpen", "Error opening input file: ");
    xcpt.desc_cat (infile);
    throw xcpt;
  }

  while (ifs_student) add_student(ifs_student);

  ifs_student.close();

  cout << "\nProblem created.\n";

  return problem();

}


//======================================================================
//                           License
//
//   Permission is hereby granted, free of charge, to any 
//   person obtaining a copy of this software and associated 
//   documentation files (the "Software"), to deal in the Software 
//   without restriction, including without limitation the rights 
//   to use, copy, modify, merge, publish, distribute, sublicense, 
//   and/or sell copies of the Software, and to permit persons to 
//   whom the Software is furnished to do so, subject to the 
//   following conditions:
//
//   The copyright notice and this license shall be included in all 
//   copies or substantial portions of the Software.
//
//   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
//   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES 
//   OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
//   NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
//   HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
//   WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
//   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
//   OTHER DEALINGS IN THE SOFTWARE.
//
//
//**********************************************************************

