//**********************************************************************
//  roomprob.cxx   -   Member functions for roomprob.h
//
//  Copyright Brian Marshall 1999
//
//  See the license at end of this file.
//
//  Brian Marshall    bmarshal@agt.net    (403)651-0584
//
//  99/05/27  Brian Marshall  development begun.
//
//
//----------------------------------------------------------------------

#include "roomprob.h"

//======================================================================
//  rmRoomAspect   -   Factory to a room.
//
//----------------------------------------------------------------------
//  Take note of an event.

void rmRoomAspect::note (dhEvent *x) { 

  if (x->type() == dhEv_Dcsn_Made) {
    m_num_student++;
  } else if (x->type() == dhEv_Dcsn_UnMade) {
    m_num_student--;
  }

}


//----------------------------------------------------------------------
//  Evaluate the passed way.

aipGoodness rmRoomAspect::evaluate (dhWay *way) {

  short too_many = (m_num_student + 1) - m_max_student;
  if (too_many < 0) too_many = 0;
  set_goodness (aipNeutral + (too_many * aipBad));
  return goodness();

}


//======================================================================
//  rmOptAspect   -   Factory to an option.
//
//----------------------------------------------------------------------
//  Evaluate this way based on:
//
//  Note that the subject of the option should not be the subject
//  of any other chosen option in the problem of this decision.

aipGoodness rmOptAspect::evaluate (dhWay *way) {

  aipGoodness good = aipVeryGood;

  good += (m_prefer * aipVerySlightlyBad);

  dhProblem *prob = way->subprob()->problem();
  if (prob) {
     for ( int isp=0; isp<prob->num_subprob(); isp++ ) {
       dhSubProb *sp = prob->subprob(isp);
       if ( sp && sp != way->subprob() ) {
	 dhWay *the_way = sp->this_way();
	 if ( the_way && the_way->option() ) {
	   if ( way->option() == the_way->option() ) {
	     good += aipBad;
	   }
	 }
       }
     }
  }

  set_goodness (good);

  return goodness();

}


//======================================================================
//  rmFactory   -   Factory to build Room problems.
//
//----------------------------------------------------------------------
//  Construct a factory.

rmFactory::rmFactory  (void) {

  m_problem = 0;

}


//----------------------------------------------------------------------
//  Destruct a factory.

rmFactory::~rmFactory  (void) {

  if (m_problem) delete m_problem;

}


//----------------------------------------------------------------------
//  Return a pointer to the problem being built (or throw an
//  exception).

rmProblem * rmFactory::problem  (void) {

  if (! m_problem ) {
    throw aipXcpt ("rmFctryNoProb", 
		   "Request for nonexistent problem.");
  }

  return m_problem;

}


//----------------------------------------------------------------------
//  Return a pointer to the last room added to the problem (or throw 
//  an exception).


rmRoom * rmFactory::added_room (void) {

  short num_room = problem()->num_room();
  if (num_room < 1) {
    throw aipXcpt ("rmFctryNoRoom", "Request for nonexistent room.");
  }

  return problem()->room(num_room-1);

}

//----------------------------------------------------------------------
//  Return a pointer to the last student added to the problem (or throw 
//  an exception).


rmStudent * rmFactory::added_student (void) {

  short num_student = problem()->num_student();
  if (num_student < 1) {
    throw aipXcpt ("rmFctryNoStudent",
		   "Request for nonexistent student.");
  }

  return problem()->student(num_student-1);

}


//----------------------------------------------------------------------
//  Begin constructing a new problem.  If this factory has been
//  constructing a problem, delete it first.


void rmFactory::begin_new_problem (void) {

  if (m_problem) {
    delete m_problem;
    m_problem = 0;
  }

  m_problem = new_problem();

  if ( ! m_problem ) {
    throw aipXcpt ("rmProbAllc", "Error allocating rmProblem.");
  }

}


//----------------------------------------------------------------------
//  Add a new room that can hold 'max_student' students to the
//  existing problem.

void rmFactory::add_room (short max_student) {

  if (! m_problem ) {
    throw aipXcpt ("rmAddRoomNoProb",
		   "Adding room to nonexistent problem.");
  }

  rmRoom *room = new_room();

  if ( ! room ) {
    throw aipXcpt ("rmRoomAllc", "Error allocating Room.");
  }

  m_problem->add_room(room);

  rmRoomAspect *aspect = new rmRoomAspect(max_student);

  if ( ! aspect ) {
    throw aipXcpt ("dhRoomAspectAllc",
		   "Error allocating room aspect.");
  }

  room->add_problem_aspect(aspect);

}


//----------------------------------------------------------------------
//  Add a new student that must be assigned to 'num_decision' rooms 
//  to the existing problem.  The student can have up to 'max_option' 
//  room-options in each decision.

void rmFactory::add_student (short num_decision, short max_option) {

  if (! m_problem ) {
    throw aipXcpt ("rmAddStudentNoProb",
		   "Adding student to nonexistent problem.");
  }

  rmStudent *student = new_student();

  if ( ! student ) {
    throw aipXcpt ("rmStudentAllc", "Error allocating Student.");
  }

  m_problem->add_student(student);

  student->set_max_subprob(num_decision);

  for (short idcsn=0; idcsn<num_decision; idcsn++) {

    dhDecision *decision = new dhDecision;

    if ( ! decision ) {
      throw aipXcpt ("dhDcsnAllc", "Error allocating decision.");
    }

    student->add_subprob(decision);

    decision->set_max_option(max_option);

  }   // End of loop through decisions for this student.

}


//----------------------------------------------------------------------
//  Add an option to the last student added to the problem.  That
//  student may optionally be assigned to the 'iroom'th room
//  (starting from zero) in each decision.

void  rmFactory::add_option (short iroom) {

  rmStudent *student = added_student();

  for (short isp=0; isp<student->num_subprob(); isp++) {
    dhDecision *decision = (dhDecision*)student->subprob(isp);

    dhOption *option = new dhOption;

    if ( ! option ) {
      throw aipXcpt ("dhOptAllc", "Error allocating option.");
    }

    decision->add_option(option);

    option->set_subject ( problem()->room(iroom) );

    rmOptAspect *aspect = new rmOptAspect();

    if ( ! aspect ) {
      throw aipXcpt ("dhOptAspectAllc",
		     "Error allocating opt aspect.");
    }

    option->add_problem_aspect(aspect);

    aspect->set_prefer ( decision->num_option() );

  }   // End of loop through decisions for the added student.

}


//----------------------------------------------------------------------
//  Deliver the constructed problem.

rmProblem *  rmFactory::deliver (void) {

  rmProblem *prob = problem();
  m_problem = 0;
  return prob;

}


//======================================================================
//                           License
//
//   Permission is hereby granted, free of charge, to any 
//   person obtaining a copy of this software and associated 
//   documentation files (the "Software"), to deal in the Software 
//   without restriction, including without limitation the rights 
//   to use, copy, modify, merge, publish, distribute, sublicense, 
//   and/or sell copies of the Software, and to permit persons to 
//   whom the Software is furnished to do so, subject to the 
//   following conditions:
//
//   The copyright notice and this license shall be included in all 
//   copies or substantial portions of the Software.
//
//   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
//   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES 
//   OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
//   NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
//   HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
//   WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
//   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
//   OTHER DEALINGS IN THE SOFTWARE.
//
//======================================================================

