//**********************************************************************
//  roomprob.h   -   Assign students to rooms.
//
//  This is a specialization of  dirthard.h  to solve 'room' problems.
//
//  This header file declares...
//    - dirty-hard problems, decisions, options and aspects
//    - aspects, ways, strategies, factories
//
//  This software throws aipXcpt exceptions on problems.
//
//  Copyright Brian Marshall 1999
//
//  See the license at end of this file.
//
//  Brian Marshall    bmarshal@agt.net    (403)651-0584
//
//  99/05/21  Brian Marshall  development begun.
//
//----------------------------------------------------------------------
//  The Problem Domain and Interface
//
//  Given:
//
//    -  some rooms, each with an upper limit on the number of 
//       students it can hold, and,
//
//    -  some students that each have to be assigned to one or 
//       more rooms and that each have preferences,
//
//  assign each student to the required number of rooms.
//
//  The primary purpose of this application is to assign students to
//  optional-classes.  Unfortunately, the words 'option' and 'class'
//  both have special meanings, so we went with the word 'room'.
//
//  The number of rooms to which each student is assigned, and the
//  number of preferences, can vary from student to student.  The 
//  number of students that each room can hold can vary from room 
//  to room.
//
//
//----------------------------------------------------------------------

#ifndef roomprob_h_guard
#define roomprob_h_guard

#include "dirthard.h"


//----------------------------------------------------------------------
//  Classes.

class  rmRoomAspect;       // Aspect of  a subject.
class  rmOptAspect;        // Aspect of an option.

class  rmStudent;          // Student:  a  dhProblem with decisions.
class  rmRoom;             // Room:     a  dhSubject.
class  rmProblem;          // Problem:  a  dhProblem with students.

class  rmFactory;          // Factory for building a problem.


//======================================================================
//  Aspects.
//
//----------------------------------------------------------------------
//  rmRoomAspect   -    Aspect of a room (a subject).

class rmRoomAspect : public dhAspect {

  short  m_max_student;     // Max students that can be assigned.
  short  m_num_student;     // Num students currently assigned.

public:

  rmRoomAspect (short max_student) { 
    m_max_student = max_student;
    m_num_student = 0; 
  }

  virtual ~rmRoomAspect (void) {}

  virtual void note (dhEvent *x);

  short max_student (void) const { return m_max_student; }
  short num_student (void) const { return m_num_student; }

  void set_num_student (short x) { m_num_student = x;  }

  virtual aipGoodness evaluate (dhWay *way);

};


//----------------------------------------------------------------------
//  rmOptAspect   -   Aspect of an option.
//
//  The goodness of a solution is determined here.  Room problems are
//  very clean in that they have only one across-the-board aspect that
//  determines solution goodness: the degree to which each student
//  is assigned to rooms that are preferred.


class rmOptAspect : public dhAspect {

  short  m_prefer;    // 0 = first choice, 1 = next choice, etc.

 public:

  rmOptAspect (void) { m_prefer = 0; }
  virtual ~rmOptAspect (void) {}

  virtual aipGoodness evaluate (dhWay *way);

  short prefer (void) const { return m_prefer; }

  void set_prefer (short x) { m_prefer = x; }

};


//======================================================================
//  rmRoom   -   Room (a subject).

class rmRoom : public dhSubject {

public:

};


//======================================================================
//  rmStudent   -   Student (a problem).
//
//  A student that needs to be assigned to 3 rooms is a problem
//  with 3 decisions.
//
//  The options added to each decision must be added in order
//  of decreasing preference.
//
//  Calling  room(1)  after the problem has been solved will return 
//  a pointer to the 2nd room assigned in the best solution.  The
//  argument is like an index into an array of assigned rooms.

class rmStudent : public dhProblem {

public:

  rmRoom * room (short iroom) const { 
    return (rmRoom*) subprob(iroom)->best_way()->subject();
  }

};


//======================================================================
//  rmProblem   -   Room problem.
//
//  A room-problem with 200 students and 15 rooms is problem with 
//  15 subjects and 200 subproblems.

class rmProblem : public dhProblem {

public:

  short num_student (void) const { return num_subprob(); }
  short num_room    (void) const { return num_subject();  }

  rmRoom * room (short iroom) const {
    return (rmRoom*) subject(iroom);
  }

  rmStudent * student (short istudent) const { 
    return (rmStudent*) subprob(istudent);
  }

  void add_room    (rmRoom    *x) { add_subject(x); }
  void add_student (rmStudent *x) { add_subprob(x); }

};


//======================================================================
//  rmFactory   -   Factory for building room problems.
//
//  A room problem is constructed by asking the factory to:
//   - begin constructing a problem, deleting an existing one,
//     if one exists
//   - add some rooms to the problem
//   - add some students and options to the problem
//   - deliver the problem
//
//  Options are added, in decreasing order of preference, to each 
//  student after the student is added to the problem.  Options take 
//  the form "The last student added has chosen the 'iroom'th room 
//  as his/her next choice."
//
//  If a factory is destroyed with an undelivered problem, the
//  problem is deleted.
//
//  Subclasses may -
//   - override the new_xxx() functions to allocate object of a
//     subclass of an rmProblem, rmRoom or rmStudent.
//   - use added_room() and added_student() to get a pointer to
//     the most recently added object so that if may be configured.
//   - use problem() to get a pointer to the problem being built
//     if it is necessary to search/inspect it for something.
//
//  Note that added_room(), added_student() and problem() will
//  throw an exception if they can't return a pointer to a valid
//  object.


class  rmFactory : public dhFactory {

  rmProblem   * m_problem;   // Problem being constructed.

 protected:

  virtual rmProblem  * new_problem  (void) { return new rmProblem; }
  virtual rmRoom     * new_room     (void) { return new rmRoom;    }
  virtual rmStudent  * new_student  (void) { return new rmStudent; }

  rmProblem * problem       (void);
  rmRoom    * added_room    (void);
  rmStudent * added_student (void);

 public:

  rmFactory (void);
  virtual ~rmFactory (void);

  void begin_new_problem (void);

  void  add_room     (short max_student);
  void  add_student  (short num_decision, short max_option);

  void add_option (short iroom);  // Last student's next choice.

  rmProblem * deliver (void);

};


//======================================================================

#endif


//----------------------------------------------------------------------
//                           License
//
//   Permission is hereby granted, free of charge, to any 
//   person obtaining a copy of this software and associated 
//   documentation files (the "Software"), to deal in the Software 
//   without restriction, including without limitation the rights 
//   to use, copy, modify, merge, publish, distribute, sublicense, 
//   and/or sell copies of the Software, and to permit persons to 
//   whom the Software is furnished to do so, subject to the 
//   following conditions:
//
//   The copyright notice and this license shall be included in all 
//   copies or substantial portions of the Software.
//
//   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
//   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES 
//   OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
//   NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
//   HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
//   WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
//   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
//   OTHER DEALINGS IN THE SOFTWARE.
//
//**********************************************************************

